from typing import TYPE_CHECKING, Any, TypeVar, Union

from edgy.core.db.fields.foreign_keys import ForeignKey
from edgy.core.terminal import Print

if TYPE_CHECKING:
    from edgy import Model
    from edgy.core.db.fields.base import BaseField

T = TypeVar("T", bound="Model")

terminal = Print()

class OneToOneField(ForeignKey):
    """
    Representation of a one to one field.
    """

    def __new__(  # type: ignore
        cls,
        to: Union["Model", str],
        **kwargs: Any,
    ) -> "BaseField":
        for argument in ["index", "unique"]:
            if argument in kwargs:
                terminal.write_warning(f"Declaring {argument} on a OneToOneField has no effect.")
        kwargs["unique"] = True

        return super().__new__(cls, to=to, **kwargs)

OneToOne = OneToOneField



from typing import TYPE_CHECKING, Optional, Sequence, Type, TypeVar, Union, cast

import sqlalchemy

from edgy.core.connection.database import Database
from edgy.core.db.context_vars import set_queryset_database, set_queryset_schema, set_schema

if TYPE_CHECKING:
    from edgy import QuerySet, Registry
    from edgy.core.db.models import Model, ReflectModel


_EdgyModel = TypeVar("_EdgyModel", bound="Model")
ReflectEdgyModel = TypeVar("ReflectEdgyModel", bound="ReflectModel")

EdgyModel = Union[_EdgyModel, ReflectEdgyModel]


class QuerySetPropsMixin:
    """
    Properties used by the Queryset are placed in isolation
    for clean access and maintainance.
    """

    @property
    def database(self) -> Database:
        if getattr(self, "_database", None) is None:
            return cast("Database", self.model_class.meta.registry.database)
        return self._database

    @database.setter
    def database(self, value: Database) -> None:
        self._database = value

    @property
    def table(self) -> sqlalchemy.Table:
        if getattr(self, "_table", None) is None:
            return cast("sqlalchemy.Table", self.model_class.table)
        return self._table

    @table.setter
    def table(self, value: sqlalchemy.Table) -> None:
        self._table = value

    @property
    def pknames(self) -> Sequence[str]:
        return self.model_class.pknames  # type: ignore

    @property
    def pkcolumns(self) -> Sequence[str]:
        return self.model_class.pkcolumns  # type: ignore


class TenancyMixin:
    """
    Mixin used for querying a possible multi tenancy application
    """

    def using(self, schema: str) -> "QuerySet":
        """
        Enables and switches the db schema.

        Generates the registry object pointing to the desired schema
        using the same connection.
        """
        queryset = set_queryset_schema(self, self.model_class, value=schema)
        return queryset

    def using_with_db(self, connection_name: str, schema: Optional[str] = None) -> "QuerySet":
        """
        Enables and switches the database connection.

        Generates the new queryset using the selected connection provided in the extra of the model
        registry.
        """
        assert (
            connection_name in self.model_class.meta.registry.extra
        ), f"`{connection_name}` is not in the connections extra of the model`{self.model_class.__name__}` registry"

        connection: Type["Registry"] = self.model_class.meta.registry.extra[connection_name]
        if schema:
            return set_queryset_database(self, self.model_class, connection, schema)
        queryset = set_queryset_database(self, self.model_class, connection)
        return queryset


def activate_schema(tenant_name: str) -> None:
    """
    Activates the tenant for the context of the query.
    """
    set_schema(tenant_name)


def deactivate_schema() -> None:
    """
    Deactivates the tenant for the context of the query.
    """
    set_schema(None)



class DescriptiveMeta:
    """
    The `Meta` class used to configure each metadata of the model.
    Abstract classes are not generated in the database, instead, they are simply used as
    a reference for field generation.

    Usage:

    .. code-block:: python3

        class User(Model):
            ...

            class Meta:
                registry = models
                tablename = "users"

    """

    ...  # pragma: no cover



from typing import Any, Optional, Tuple, Type, Union

from edgy.core.db.models.metaclasses import (
    BaseModelMeta,
    MetaInfo,
    get_model_registry,
)
from edgy.exceptions import ImproperlyConfigured


def _check_model_inherited_tenancy(bases: Tuple[Type, ...]) -> Union[bool, None]:
    """
    When a registry is missing from the Meta class, it should look up for the bases
    and obtain the first found registry.

    If not found, then a ImproperlyConfigured exception is raised.
    """
    is_tenant: Optional[bool] = None

    for base in bases:
        meta: MetaInfo = getattr(base, "meta", None)  # type: ignore
        if not meta:
            continue

        meta_tenant: Optional[bool] = getattr(meta, "is_tenant", None)
        if meta_tenant is not None and meta_tenant is not False:
            is_tenant = meta_tenant
            break

    return is_tenant


class TenantMeta(MetaInfo):
    def __init__(self, meta: Any = None, **kwargs: Any) -> None:
        super().__init__(meta, **kwargs)
        self.is_tenant: bool = getattr(meta, "is_tenant", False)

    def set_tenant(self, is_tenant: bool) -> None:
        self.is_tenant = is_tenant


class BaseTenantMeta(BaseModelMeta):
    """
    The metaclass for the base tenant used by the Edgy contrib mixin.

    This should only be used by the contrib or if you decided to use
    your own tenant model using the `is_tenant` inside the `Meta` object.
    """

    def __new__(cls, name: str, bases: Tuple[Type, ...], attrs: Any) -> Any:
        meta_class: object = attrs.get("Meta", type("Meta", (), {}))
        new_model = super().__new__(cls, name, bases, attrs)
        meta: TenantMeta = TenantMeta(new_model.meta)

        if hasattr(meta_class, "is_tenant"):
            meta.set_tenant(meta_class.is_tenant)

        # Handle the registry of models
        if getattr(meta, "registry", None) is None:
            if hasattr(new_model, "__db_model__") and new_model.__db_model__:
                meta.registry = get_model_registry(bases)
            else:
                return new_model
        registry = meta.registry
        if registry is None:
            raise ImproperlyConfigured(
                "Registry for the table not found in the Meta class or any of the superclasses. You must set the registry in the Meta."
            )

        new_model.meta = meta

        # Check if is tenant
        is_tenant = _check_model_inherited_tenancy(bases)
        if is_tenant:
            new_model.meta.is_tenant = is_tenant

        if registry:
            try:
                if meta.is_tenant and not meta.abstract:
                    registry.tenant_models[new_model.__name__] = new_model
            except KeyError:
                ...  # pragma: no cover
        return new_model



from typing import Any

from pydantic import BaseModel


class HashableBaseModel(BaseModel):
    """
    Pydantic BaseModel by default doesn't handle with hashable types the same way
    a python object would and therefore there are types that are mutable (list, set)
    not hashable and those need to be handled properly.

    HashableBaseModel handles those corner cases.
    """

    __slots__ = ["__weakref__"]

    def __hash__(self) -> Any:
        values: Any = {}
        for key, value in self.__dict__.items():
            values[key] = None
            if isinstance(value, (list, set)):
                values[key] = tuple(value)
            else:
                values[key] = value
        return hash((type(self),) + tuple(values))


class ArbitraryHashableBaseModel(HashableBaseModel):
    """
    Same as HashableBaseModel but allowing arbitrary values
    """

    class Config:
        extra = "allow"
        arbitrary_types_allowed = True



from typing import Any, Dict

from edgy.core.connection.database import Database
from edgy.core.connection.registry import Registry


class TenantRegistry(Registry):
    def __init__(self, database: Database, **kwargs: Any) -> None:
        super().__init__(database, **kwargs)
        self.tenant_models: Dict[str, Any] = {}



import copy
import decimal
from functools import cached_property
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    FrozenSet,
    Literal,
    Optional,
    Pattern,
    Sequence,
    Tuple,
    Type,
    Union,
    cast,
)

import sqlalchemy
from pydantic import BaseModel
from pydantic.fields import FieldInfo

from edgy.core.connection.registry import Registry
from edgy.types import Undefined

if TYPE_CHECKING:
    from edgy import Model, ReflectModel

def _removesuffix(text: str, suffix: str) -> str:
    # TODO: replace with _removesuffix when python3.9 is minimum
    if text.endswith(suffix):
        return text[: -len(suffix)]
    else:
        return text


class BaseField(FieldInfo):
    """
    The base field for all Edgy data model fields.
    """

    __namespace__: ClassVar[Union[Dict[str, Any], None]] = None

    def __init__(
        self,
        *,
        default: Any = Undefined,
        server_default: Any = Undefined,
        inherit: bool = True,
        **kwargs: Any,
    ) -> None:
        self.max_digits: str = kwargs.pop("max_digits", None)
        self.decimal_places: str = kwargs.pop("decimal_places", None)
        self.server_default: Any = server_default
        self.read_only: bool = kwargs.pop("read_only", False)
        self.primary_key: bool = kwargs.pop("primary_key", False)
        self.autoincrement: bool = kwargs.pop("autoincrement", False)
        self.inherit = inherit

        super().__init__(**kwargs)

        self.null: bool = kwargs.pop("null", False)
        if self.null and default is Undefined:
            default = None
        if default is not Undefined:
            self.default = default
        if (default is not None and default is not Undefined) or (
            self.server_default is not None and self.server_default != Undefined
        ):
            self.null = True
        self.field_type: Any = kwargs.pop("__type__", None)
        self.__original_type__: type = kwargs.pop("__original_type__", None)
        self.column_type: Optional[Any] = kwargs.pop("column_type", None)
        self.constraints: Sequence[sqlalchemy.Constraint] = kwargs.pop("constraints", [])
        self.skip_absorption_check: bool = kwargs.pop("skip_absorption_check", False)
        self.help_text: Optional[str] = kwargs.pop("help_text", None)
        self.pattern: Pattern = kwargs.pop("pattern", None)
        self.unique: bool = kwargs.pop("unique", False)
        self.index: bool = kwargs.pop("index", False)
        self.choices: Sequence = kwargs.pop("choices", [])
        self.owner: Union[Type["Model"], Type["ReflectModel"]] = kwargs.pop("owner", None)
        # field name, set when retrieving
        self.name: str = kwargs.get("name", None)
        self.alias: str = kwargs.pop("name", None)
        self.regex: str = kwargs.pop("regex", None)
        self.format: str = kwargs.pop("format", None)
        self.min_length: Optional[int] = kwargs.pop("min_length", None)
        self.max_length: Optional[int] = kwargs.pop("max_length", None)
        self.minimum: Optional[Union[int, float, decimal.Decimal]] = kwargs.pop("minimum", None)
        self.maximum: Optional[Union[int, float, decimal.Decimal]] = kwargs.pop("maximum", None)
        self.multiple_of: Optional[Union[int, float, decimal.Decimal]] = kwargs.pop("multiple_of", None)
        self.server_onupdate: Any = kwargs.pop("server_onupdate", None)
        self.registry: Registry = kwargs.pop("registry", None)
        self.comment: str = kwargs.pop("comment", None)
        self.secret: bool = kwargs.pop("secret", False)


        # set remaining attributes
        for name, value in kwargs.items():
            setattr(self, name, value)

        if self.primary_key:
            self.field_type = Any
            self.null = True

        if isinstance(self.default, bool):
            self.null = True
        self.__namespace__ = {k: v for k, v in self.__dict__.items() if k != "__namespace__"}

    @property
    def namespace(self) -> Any:
        """Returns the properties added to the fields in a dict format"""
        return self.__namespace__

    def is_required(self) -> bool:
        """Check if the argument is required.

        Returns:
            `True` if the argument is required, `False` otherwise.
        """
        if self.primary_key:
            if self.autoincrement:
                return False
        return False if self.null or self.server_default else True

    def get_alias(self) -> str:
        """
        Used to translate the model column names into database column tables.
        """
        return self.name

    def has_default(self) -> bool:
        """Checks if the field has a default value set"""
        return bool(self.default is not None and self.default is not Undefined)

    def get_columns(self, name: str) -> Sequence[sqlalchemy.Column]:
        """
        Returns the columns of the field being declared.
        """
        return []

    def get_column_names(self, name: str="") -> FrozenSet[str]:
        if name:
            return self.owner.meta.field_to_column_names[name]
        return self.owner.meta.field_to_column_names[self.name]

    def clean(self, field_name: str, value: Any, for_query: bool=False) -> Dict[str, Any]:
        """
        Validates a value and transform it into columns which can be used for querying and saving.
        for_query: is used for querying. Should have all columns used for querying set.
        """
        return {}

    def to_model(self, field_name: str, value: Any, phase: str = "") -> Dict[str, Any]:
        """
        Inverse of clean. Transforms column(s) to a field for a pydantic model (EdgyBaseModel).
        Validation happens later.
        """
        return {field_name: value}

    def get_embedded_fields(self, field_name: str, fields_mapping: Dict[str, "BaseField"]) -> Dict[str, "BaseField"]:
        """
        Define extra fields on the fly. Often no owner is available yet.

        Arguments are:
        name: the field name
        fields_mapping: the existing fields

        Note: the returned fields are changed after return, so you should
              return new fields or copies. Also set the owner of the field to them before returning
        """
        return {}

    def embed_field(self, prefix: str, new_fieldname:str, owner: Optional[Union[Type["Model"], Type["ReflectModel"]]]=None, parent: Optional["BaseField"]=None) -> Optional["BaseField"]:
        """
        Embed this field or return None to prevent embedding.
        Must return a copy with name and owner set when not returning None.
        """
        field_copy = copy.copy(self)
        field_copy.name = new_fieldname
        field_copy.owner = owner  # type: ignore
        return field_copy

    def get_constraints(self) -> Any:
        return self.constraints

    def get_global_constraints(self, name: str, columns: Sequence[sqlalchemy.Column]) -> Sequence[sqlalchemy.Constraint]:
        """Return global constraints and indexes.
        Useful for multicolumn fields
        """
        return []

    def get_default_value(self) -> Any:
        # single default
        default = getattr(self, "default", None)
        if callable(default):
            return default()
        return default

    def get_default_values(self, field_name: str, cleaned_data: Dict[str, Any]) -> Any:
        # for multidefaults overwrite in subclasses get_default_values to
        # parse default values differently
        # NOTE: multi value fields should always check here if defaults were already applied
        # NOTE: when build meta fields without columns this should be empty
        if field_name in cleaned_data:
            return {}
        return {field_name: self.get_default_value()}


class Field(BaseField):
    # defines compatibility fallbacks check and get_column

    def check(self, value: Any) -> Any:
        """
        Runs the checks for the fields being validated. Single Column.
        """
        return value

    def clean(self, name: str, value: Any, for_query: bool=False) -> Dict[str, Any]:
        """
        Runs the checks for the fields being validated. Multiple columns possible
        """
        return {name: self.check(value)}

    def get_column(self, name: str) -> Optional[sqlalchemy.Column]:
        """
        Return a single column for the field declared. Return None for meta fields.
        """
        constraints = self.get_constraints()
        return sqlalchemy.Column(
            name,
            self.column_type,
            *constraints,
            primary_key=self.primary_key,
            autoincrement=self.autoincrement,
            nullable=self.null,
            index=self.index,
            unique=self.unique,
            default=self.default,
            comment=self.comment,
            server_default=self.server_default,
            server_onupdate=self.server_onupdate,
        )

    def get_columns(self, name: str) -> Sequence[sqlalchemy.Column]:
        column = self.get_column(name)
        if column is None:
            return []
        return [column]


class BaseCompositeField(BaseField):
    def translate_name(self, name: str) -> str:
        """translate name for inner objects and parsing values"""
        return name

    def get_composite_fields(self) -> Dict[str, BaseField]:
        """return dictionary of fields with untranslated names"""
        raise NotImplementedError()

    @cached_property
    def composite_fields(self) -> Dict[str, BaseField]:
        # return untranslated names
        return self.get_composite_fields()

    def clean(self, field_name: str, value: Any, for_query: bool=False) -> Dict[str, Any]:
        """
        Runs the checks for the fields being validated.
        """
        prefix = _removesuffix(field_name, self.name)
        result = {}
        if isinstance(value, dict):
            for sub_name, field in self.composite_fields.items():
                translated_name = self.translate_name(sub_name)
                if translated_name not in value:
                    if field.has_default() or not field.is_required():
                        continue
                    raise ValueError(f"Missing key: {sub_name} for {field_name}")
                result.update(field.clean(f"{prefix}{sub_name}", value[translated_name], for_query=for_query))
        else:
            for sub_name, field in self.composite_fields.items():
                translated_name = self.translate_name(sub_name)
                if not hasattr(value, translated_name):
                    if not for_query:
                        if field.has_default() or not field.is_required():
                            continue
                    raise ValueError(f"Missing attribute: {translated_name} for {field_name}")
                result.update(field.clean(f"{prefix}{sub_name}", getattr(value, translated_name), for_query=for_query))
        return result

    def to_model(self, field_name: str, value: Any, phase: str = "") -> Dict[str, Any]:
        """
        Runs the checks for the fields being validated.
        """
        result = {}
        if isinstance(value, dict):
            for sub_name, field in self.composite_fields.items():
                translated_name = self.translate_name(sub_name)
                if translated_name not in value:
                    continue
                result.update(field.to_model(sub_name, value.get(translated_name, None), phase=phase))
        else:
            for sub_name, field in self.composite_fields.items():
                translated_name = self.translate_name(sub_name)
                if not hasattr(value, translated_name):
                    continue
                result.update(field.to_model(sub_name, getattr(value, translated_name, None), phase=phase))

        return result

    def get_default_values(self, field_name: str, cleaned_data: Dict[str, Any]) -> Any:
        # fields should provide their own default which is used as long as they are in the fields_mapping
        return {}


class RelationshipField(BaseField):
    def traverse_field(self, path: str) -> Tuple[Any, str, str]:
        raise NotImplementedError()

    def is_cross_db(self) -> bool:
        raise NotImplementedError()


class PKField(BaseCompositeField):
    """
    Field for pk
    """

    def __init__(self, **kwargs: Any):
        kwargs["default"] = kwargs["server_default"] = None
        kwargs["__type__"] = kwargs["annotation"] = Any
        return super().__init__(
            **kwargs,
        )

    def __get__(self, instance: "Model", owner: Any = None) -> Union[Dict[str, Any], Any]:
        pkcolumns = cast(Sequence[str], self.owner.pkcolumns)
        pknames = cast(Sequence[str], self.owner.pknames)
        assert len(pkcolumns) >= 1
        if len(pknames) == 1:
            return getattr(instance, pknames[0], None)
        d = {}
        for key in pknames:
            translated_name = self.translate_name(key)
            field = instance.meta.fields_mapping.get(key)
            if field and hasattr(field, "__get__"):
                d[translated_name] = field.__get__(instance, owner)
            else:
                d[translated_name] = getattr(instance, key, None)
        for key in self.fieldless_pkcolumns:
            translated_name = self.translate_name(key)
            d[translated_name] = getattr(instance, key, None)
        return d

    def modify_input(self, name: str, kwargs: Dict[str, Any]) -> None:
        if name not in kwargs:
            return
        # check for collisions
        for pkname in self.owner.pknames:
            if pkname in kwargs:
                raise ValueError("Cannot specify a primary key field and the primary key itself")

    def embed_field(self, prefix: str, new_fieldname:str, owner: Optional[Union[Type["Model"], Type["ReflectModel"]]]=None, parent: Optional["BaseField"]=None) -> Optional[BaseField]:
        return None

    def clean(self, field_name: str, value: Any, for_query: bool = False) -> Dict[str, Any]:
        pknames = cast(Sequence[str], self.owner.pknames)
        pkcolumns = cast(Sequence[str], self.owner.pkcolumns)
        prefix = _removesuffix(field_name, self.name)
        assert len(pkcolumns) >= 1
        if (
            len(pknames) == 1
            and not self.is_incomplete
            and not isinstance(value, (dict, BaseModel))
        ):
            pkname = pknames[0]
            field = self.owner.meta.fields_mapping[pkname]
            return field.clean(f"{prefix}{pkname}", value, for_query=for_query)
        retdict =  super().clean(field_name, value, for_query=for_query)
        if self.is_incomplete:
            if isinstance(value, dict):
                for column_name in self.fieldless_pkcolumns:
                    translated_name = self.translate_name(column_name)
                    if translated_name not in value:
                        if not for_query:
                            continue
                        raise ValueError(f"Missing key: {translated_name} for {field_name}")
                    retdict[f"{prefix}{column_name}"] = value[translated_name]
            else:
                for column_name in self.fieldless_pkcolumns:
                    translated_name = self.translate_name(column_name)
                    if not hasattr(value, translated_name):
                        if not for_query:
                            continue
                        raise ValueError(f"Missing attribute: {translated_name} for {field_name}")
                    retdict[f"{prefix}{column_name}"] = getattr(value, translated_name)

        return retdict

    def to_model(self, field_name: str, value: Any, phase: str = "") -> Dict[str, Any]:
        pknames = cast(Sequence[str], self.owner.pknames)
        assert len(cast(Sequence[str], self.owner.pkcolumns)) >= 1
        if self.is_incomplete:
            raise ValueError("Cannot set an incomplete defined pk!")
        if (len(pknames) == 1
            # we first only redirect both
            and not isinstance(value, (dict, BaseModel))
        ):
            field = self.owner.meta.fields_mapping[pknames[0]]
            return field.to_model(pknames[0], value, phase=phase)
        return super().to_model(field_name, value, phase=phase)

    def get_composite_fields(self) -> Dict[str, BaseField]:
        return {field: self.owner.meta.fields_mapping[field] for field in cast(Sequence[str], self.owner.pknames)}

    @cached_property
    def fieldless_pkcolumns(self) -> FrozenSet[str]:
        field_less = set()
        for colname in self.owner.pkcolumns:
            if colname not in self.owner.meta.columns_to_field:
                field_less.add(colname)
        return frozenset(field_less)

    @property
    def is_incomplete(self) -> bool:
        return bool(self.fieldless_pkcolumns)


    def is_required(self) -> bool:
        return False


class BaseForeignKey(RelationshipField):
    is_m2m: bool = False

    def __init__(
        self,
        *,
        related_name: Union[str, Literal[False]] = "",
        reverse_name: str = "",
        **kwargs: Any,
    ) -> None:
        self.related_name = related_name
        # name used for backward relations
        # only useful if related_name = False because otherwise it gets overwritten
        self.reverse_name = reverse_name
        super().__init__(**kwargs)

    @property
    def target(self) -> Any:
        """
        The target of the ForeignKey model.
        """
        if not hasattr(self, "_target"):
            if isinstance(self.to, str):
                self._target = self.registry.models[self.to]  # type: ignore
            else:
                self._target = self.to
        return self._target

    @target.setter
    def target(self, value: Any) -> None:
        self._target = value

    @target.deleter
    def target(self, value: Any) -> None:
        try:
            delattr(self, "_target")
        except AttributeError:
            pass

    def is_cross_db(self) -> bool:
        return self.owner.meta.registry is not self.target.meta.registry

    def expand_relationship(self, value: Any) -> Any:
        """
        Returns the related object or the relationship object
        """
        return value

    def to_model(self, field_name: str, value: Any, phase: str = "") -> Dict[str, Any]:
        return {field_name: self.expand_relationship(value)}



import datetime
import decimal
import enum
import ipaddress
import re
import uuid
from enum import EnumMeta
from typing import (
    Any,
    Optional,
    Pattern,
    Sequence,
    Tuple,
    Union,
)

import pydantic
import sqlalchemy
from pydantic import EmailStr

from edgy.core.db.fields._internal import IPAddress
from edgy.core.db.fields._validators import IPV4_REGEX, IPV6_REGEX
from edgy.core.db.fields.base import BaseField, Field
from edgy.core.db.fields.factories import FieldFactory
from edgy.exceptions import FieldDefinitionError

CLASS_DEFAULTS = ["cls", "__class__", "kwargs"]

class CharField(FieldFactory, str):
    """String field representation that constructs the Field class and populates the values"""

    _type = str

    def __new__(  # type: ignore
        cls,
        *,
        max_length: Optional[int] = 0,
        min_length: Optional[int] = None,
        regex: Union[str, Pattern] = None,
        **kwargs: Any,
    ) -> BaseField:
        if regex is None:
            regex = None
            kwargs["pattern_regex"] = None
        elif isinstance(regex, str):
            regex = regex
            kwargs["pattern_regex"] = re.compile(regex)
        else:
            regex = regex.pattern
            kwargs["pattern_regex"] = regex

        kwargs = {
            **kwargs,
            **{key: value for key, value in locals().items() if key not in CLASS_DEFAULTS},
        }

        return super().__new__(cls, **kwargs)

    @classmethod
    def validate(cls, **kwargs: Any) -> None:
        max_length = kwargs.get("max_length", 0)
        if max_length <= 0:
            raise FieldDefinitionError(detail=f"'max_length' is required for {cls.__name__}")

        min_length = kwargs.get("min_length")
        pattern = kwargs.get("regex")

        assert min_length is None or isinstance(min_length, int)
        assert max_length is None or isinstance(max_length, int)
        assert pattern is None or isinstance(pattern, (str, Pattern))

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> Any:
        return sqlalchemy.String(length=kwargs.get("max_length"), collation=kwargs.get("collation", None))


class TextField(FieldFactory, str):
    """String representation of a text field which means no max_length required"""

    _type = str

    def __new__(cls, **kwargs: Any) -> BaseField:  # type: ignore
        kwargs = {
            **kwargs,
            **{key: value for key, value in locals().items() if key not in CLASS_DEFAULTS},
        }
        return super().__new__(cls, **kwargs)

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> Any:
        return sqlalchemy.Text(collation=kwargs.get("collation", None))


class Number(FieldFactory):
    @classmethod
    def validate(cls, **kwargs: Any) -> None:
        minimum = kwargs.get("minimum", None)
        maximum = kwargs.get("maximum", None)

        if (minimum is not None and maximum is not None) and minimum > maximum:
            raise FieldDefinitionError(detail="'minimum' cannot be bigger than 'maximum'")


class IntegerField(Number, int):
    """
    Integer field factory that construct Field classes and populated their values.
    """

    _type = int

    def __new__(  # type: ignore
        cls,
        *,
        minimum: Optional[int] = None,
        maximum: Optional[int] = None,
        multiple_of: Optional[int] = None,
        **kwargs: Any,
    ) -> BaseField:
        if kwargs.get("primary_key", False):
            kwargs.setdefault("autoincrement", True)
        kwargs = {
            **kwargs,
            **{k: v for k, v in locals().items() if k not in ["cls", "__class__", "kwargs"]},
        }
        return super().__new__(cls, **kwargs)

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> Any:
        return sqlalchemy.Integer()


class FloatField(Number, float):
    """Representation of a int32 and int64"""

    _type = float

    def __new__(  # type: ignore
        cls,
        *,
        mininum: Optional[float] = None,
        maximun: Optional[float] = None,
        multiple_of: Optional[int] = None,
        **kwargs: Any,
    ) -> BaseField:
        kwargs = {
            **kwargs,
            **{key: value for key, value in locals().items() if key not in CLASS_DEFAULTS},
        }
        return super().__new__(cls, **kwargs)

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> Any:
        return sqlalchemy.Float()


class BigIntegerField(IntegerField):
    """Representation of big integer field"""

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> Any:
        return sqlalchemy.BigInteger()


class SmallIntegerField(IntegerField):
    """Represents a small integer field"""

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> Any:
        return sqlalchemy.SmallInteger()


class DecimalField(Number, decimal.Decimal):
    _type = decimal.Decimal

    def __new__(  # type: ignore
        cls,
        *,
        minimum: float = None,
        maximum: float = None,
        multiple_of: int = None,
        max_digits: int = None,
        decimal_places: int = None,
        **kwargs: Any,
    ) -> BaseField:
        kwargs = {
            **kwargs,
            **{k: v for k, v in locals().items() if k not in ["cls", "__class__", "kwargs"]},
        }
        return super().__new__(cls, **kwargs)

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> Any:
        return sqlalchemy.Numeric(precision=kwargs.get("max_digits"), scale=kwargs.get("decimal_places"))

    @classmethod
    def validate(cls, **kwargs: Any) -> None:
        super().validate(**kwargs)

        max_digits = kwargs.get("max_digits")
        decimal_places = kwargs.get("decimal_places")
        if max_digits is None or max_digits < 0 or decimal_places is None or decimal_places < 0:
            raise FieldDefinitionError("max_digits and decimal_places are required for DecimalField")


class BooleanField(FieldFactory, int):
    """Representation of a boolean"""

    _type = bool

    def __new__(  # type: ignore
        cls,
        *,
        default: Optional[bool] = False,
        **kwargs: Any,
    ) -> BaseField:
        kwargs = {
            **kwargs,
            **{key: value for key, value in locals().items() if key not in CLASS_DEFAULTS},
        }
        return super().__new__(cls, **kwargs)

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> Any:
        return sqlalchemy.Boolean()


class AutoNowMixin(FieldFactory):
    def __new__(  # type: ignore
        cls,
        *,
        auto_now: Optional[bool] = False,
        auto_now_add: Optional[bool] = False,
        **kwargs: Any,
    ) -> BaseField:
        if auto_now_add and auto_now:
            raise FieldDefinitionError("'auto_now' and 'auto_now_add' cannot be both True")

        if auto_now_add or auto_now:
            kwargs["read_only"] = True

        kwargs = {
            **kwargs,
            **{k: v for k, v in locals().items() if k not in CLASS_DEFAULTS},
        }
        return super().__new__(cls, **kwargs)


class DateTimeField(AutoNowMixin, datetime.datetime):
    """Representation of a datetime field"""

    _type = datetime.datetime

    def __new__(  # type: ignore
        cls,
        *,
        auto_now: Optional[bool] = False,
        auto_now_add: Optional[bool] = False,
        **kwargs: Any,
    ) -> BaseField:
        if auto_now_add or auto_now:
            kwargs["default"] = datetime.datetime.now

        kwargs = {
            **kwargs,
            **{k: v for k, v in locals().items() if k not in CLASS_DEFAULTS},
        }
        return super().__new__(cls, **kwargs)

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> Any:
        return sqlalchemy.DateTime()


class DateField(AutoNowMixin, datetime.date):
    """Representation of a date field"""

    _type = datetime.date

    def __new__(  # type: ignore
        cls,
        *,
        auto_now: Optional[bool] = False,
        auto_now_add: Optional[bool] = False,
        **kwargs: Any,
    ) -> BaseField:
        if auto_now_add or auto_now:
            kwargs["default"] = datetime.date.today

        kwargs = {
            **kwargs,
            **{k: v for k, v in locals().items() if k not in CLASS_DEFAULTS},
        }
        return super().__new__(cls, **kwargs)

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> Any:
        return sqlalchemy.Date()


class TimeField(FieldFactory, datetime.time):
    """Representation of a time field"""

    _type = datetime.time

    def __new__(cls, **kwargs: Any) -> BaseField:  # type: ignore
        kwargs = {
            **kwargs,
            **{k: v for k, v in locals().items() if k not in CLASS_DEFAULTS},
        }
        return super().__new__(cls, **kwargs)

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> Any:
        return sqlalchemy.Time()


class JSONField(FieldFactory, pydantic.Json):  # type: ignore
    """Representation of a JSONField"""

    _type = Any

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> Any:
        return sqlalchemy.JSON()


class BinaryField(FieldFactory, bytes):
    """Representation of a binary"""

    _type = bytes

    def __new__(cls, *, max_length: Optional[int] = 0, **kwargs: Any) -> BaseField:  # type: ignore
        kwargs = {
            **kwargs,
            **{k: v for k, v in locals().items() if k not in CLASS_DEFAULTS},
        }
        return super().__new__(cls, **kwargs)

    @classmethod
    def validate(cls, **kwargs: Any) -> None:
        max_length = kwargs.get("max_length", None)
        if max_length <= 0:
            raise FieldDefinitionError(detail="Parameter 'max_length' is required for BinaryField")

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> Any:
        return sqlalchemy.JSON(none_as_null=kwargs.get("sql_nullable", False))


class UUIDField(FieldFactory, uuid.UUID):
    """Representation of a uuid"""

    _type = uuid.UUID

    def __new__(cls, **kwargs: Any) -> BaseField:  # type: ignore
        kwargs = {
            **kwargs,
            **{k: v for k, v in locals().items() if k not in CLASS_DEFAULTS},
        }

        return super().__new__(cls, **kwargs)

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> Any:
        return sqlalchemy.UUID()


class ChoiceField(FieldFactory):
    """Representation of an Enum"""

    _type = enum.Enum

    def __new__(  # type: ignore
        cls,
        choices: Optional[Sequence[Union[Tuple[str, str], Tuple[str, int]]]] = None,
        **kwargs: Any,
    ) -> BaseField:
        kwargs = {
            **kwargs,
            **{k: v for k, v in locals().items() if k not in CLASS_DEFAULTS},
        }
        return super().__new__(cls, **kwargs)

    @classmethod
    def validate(cls, **kwargs: Any) -> None:
        choice_class = kwargs.get("choices")
        if choice_class is None or not isinstance(choice_class, EnumMeta):
            raise FieldDefinitionError("ChoiceField choices must be an Enum")

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> sqlalchemy.Enum:
        choice_class = kwargs.get("choices")
        return sqlalchemy.Enum(choice_class)


class PasswordField(CharField):
    """
    Representation of a Password
    """

    @classmethod
    def get_column_type(self, **kwargs: Any) -> sqlalchemy.String:
        return sqlalchemy.String(length=kwargs.get("max_length"))


class EmailField(CharField):
    _type = EmailStr

    @classmethod
    def get_column_type(self, **kwargs: Any) -> sqlalchemy.String:
        return sqlalchemy.String(length=kwargs.get("max_length"))


class URLField(CharField):
    @classmethod
    def get_column_type(self, **kwargs: Any) -> sqlalchemy.String:
        return sqlalchemy.String(length=kwargs.get("max_length"))


class _IPAddressField(Field):
    def check(self, value: Any) -> Any:
        if self.is_native_type(value):
            return value

        match_ipv4 = IPV4_REGEX.match(value)
        match_ipv6 = IPV6_REGEX.match(value)

        if not match_ipv4 and not match_ipv6:
            raise ValueError("Must be a valid IP format.")

        try:
            return ipaddress.ip_address(value)
        except ValueError:
            raise ValueError("Must be a real IP.")  # noqa


class IPAddressField(FieldFactory, str):
    _bases = (_IPAddressField,)
    _type = Union[ipaddress.IPv4Address, ipaddress.IPv6Address]

    def is_native_type(self, value: str) -> bool:
        return isinstance(value, (ipaddress.IPv4Address, ipaddress.IPv6Address))

    def __new__(  # type: ignore
        cls,
        **kwargs: Any,
    ) -> BaseField:
        kwargs = {
            **kwargs,
            **{key: value for key, value in locals().items() if key not in CLASS_DEFAULTS},
        }

        return super().__new__(cls, **kwargs)

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> IPAddress:
        return IPAddress()



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Union

import sqlalchemy
from sqlalchemy import ColumnCollection

if TYPE_CHECKING:
    from edgy.core.db.models import Model


class _EnhancedClausesHelper:
    def __init__(self, op: Any, default_empty: Any) -> None:
        self.op = op
        self.default_empty = default_empty

    def __call__(self, *args: Any) -> Any:
        if len(args) == 0:
            args = (self.default_empty,)
        return self.op(*args)

    def from_kwargs(self, columns_or_model: Union[Model, ColumnCollection], /, **kwargs: Any) -> Any:
        if not isinstance(columns_or_model, ColumnCollection) and hasattr(columns_or_model, "columns"):
            columns_or_model = columns_or_model.columns
        return self.op(*(getattr(columns_or_model, item[0]) == item[1] for item in kwargs.items()))


or_ = _EnhancedClausesHelper(sqlalchemy.or_, sqlalchemy.false())
or_.__doc__ = """
    Creates a SQL Alchemy OR clause for the expressions being passed.
"""
and_ = _EnhancedClausesHelper(sqlalchemy.and_, sqlalchemy.true())
and_.__doc__ = """
    Creates a SQL Alchemy AND clause for the expressions being passed.
"""

# alias
Q = and_


def not_(clause: Any) -> Any:
    """
    Creates a SQL Alchemy NOT clause for the expressions being passed.
    """
    return sqlalchemy.not_(clause)



from functools import lru_cache
from typing import Any, Literal, Sequence, Union, cast

from edgy.core.db.constants import CASCADE, RESTRICT, SET_NULL
from edgy.core.db.fields.base import BaseField, Field
from edgy.exceptions import FieldDefinitionError

CLASS_DEFAULTS = ["cls", "__class__", "kwargs"]


class FieldFactoryMeta(type):
    def __instancecheck__(self, instance: Any) -> bool:
        return super().__instancecheck__(instance) or isinstance(instance, self._get_field_cls(self))

    def __subclasscheck__(self, subclass: Any) -> bool:
        return super().__subclasscheck__(subclass) or issubclass(subclass, self._get_field_cls(self))


class FieldFactory(metaclass=FieldFactoryMeta):
    """The base for all model fields to be used with Edgy"""

    _bases: Sequence[Any] = (Field,)
    _type: Any = None

    def __new__(cls, **kwargs: Any) -> BaseField:
        cls.validate(**kwargs)
        return cls.build_field(**kwargs)

    @classmethod
    def build_field(cls, **kwargs: Any) -> BaseField:
        column_type = cls.get_column_type(**kwargs)
        pydantic_type = cls.get_pydantic_type(**kwargs)
        constraints = cls.get_constraints(**kwargs)
        default: None = kwargs.pop("default", None)
        server_default: None = kwargs.pop("server_default", None)

        new_field = cls._get_field_cls(cls)
        return new_field(  # type: ignore
            __type__=pydantic_type,
            annotation=pydantic_type,
            column_type=column_type,
            default=default,
            server_default=server_default,
            constraints=constraints,
            **kwargs,
        )

    @classmethod
    def validate(cls, **kwargs: Any) -> None:  # pragma no cover
        """
        Used to validate if all required parameters on a given field type are set.
        :param kwargs: all params passed during construction
        :type kwargs: Any
        """

    @classmethod
    def get_constraints(cls, **kwargs: Any) -> Sequence[Any]:
        """Returns the propery column type for the field, None for Metafields"""
        return []

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> Any:
        """Returns the propery column type for the field, None for Metafields"""
        return None

    @classmethod
    def get_pydantic_type(cls, **kwargs: Any) -> Any:
        """Returns the type for pydantic"""
        return cls._type

    @staticmethod
    @lru_cache(None)
    def _get_field_cls(cls: "FieldFactory") -> BaseField:
        return cast(BaseField, type(cls.__name__, cast(Any, cls._bases), {}))


class ForeignKeyFieldFactory(FieldFactory):
    """The base for all model fields to be used with Edgy"""

    _type: Any = Any

    def __new__(
        cls,
        *,
        to: Any = None,
        null: bool = False,
        on_update: str = CASCADE,
        on_delete: str = RESTRICT,
        related_name: Union[str, Literal[False]] = "",
        server_onupdate: Any = None,
        default: Any = None,
        server_default: Any = None,
        **kwargs: Any,
    ) -> BaseField:
        kwargs = {
            **kwargs,
            **{key: value for key, value in locals().items() if key not in CLASS_DEFAULTS},
        }

        cls.validate(**kwargs)
        # update related name when available
        if related_name:
            kwargs["related_name"] = related_name.lower()
        return cls.build_field(**kwargs)

    @classmethod
    def validate(cls, **kwargs: Any) -> None:
        """default validation useful for one_to_one and foreign_key"""
        on_delete = kwargs.get("on_delete", CASCADE)
        on_update = kwargs.get("on_update", RESTRICT)
        null = kwargs.get("null", False)

        if on_delete is None:
            raise FieldDefinitionError("on_delete must not be null.")

        if on_delete == SET_NULL and not null:
            raise FieldDefinitionError("When SET_NULL is enabled, null must be True.")

        if on_update and (on_update == SET_NULL and not null):
            raise FieldDefinitionError("When SET_NULL is enabled, null must be True.")
        related_name = kwargs.get("related_name", "")

        # tolerate None and False
        if related_name and not isinstance(related_name, str):
            raise FieldDefinitionError("related_name must be a string.")



import asyncio
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Sequence, Type, Union, cast

from edgy.core.db.fields.base import RelationshipField
from edgy.core.db.models.base import EdgyBaseModel
from edgy.exceptions import QuerySetError

if TYPE_CHECKING:  # pragma: no cover
    from sqlalchemy.engine.result import Row

    from edgy import Model, Prefetch, QuerySet


class ModelRow(EdgyBaseModel):
    """
    Builds a row for a specific model
    """

    class Meta:
        abstract = True

    @classmethod
    def from_sqla_row(
        cls,
        row: "Row",
        select_related: Optional[Sequence[Any]] = None,
        prefetch_related: Optional[Sequence["Prefetch"]] = None,
        is_only_fields: bool = False,
        only_fields: Sequence[str] = None,
        is_defer_fields: bool = False,
        exclude_secrets: bool = False,
        using_schema: Union[str, None] = None,
    ) -> Optional[Type["Model"]]:
        """
        Class method to convert a SQLAlchemy Row result into a EdgyModel row type.

        Looping through select_related fields if the query comes from a select_related operation.
        Validates if exists the select_related and related_field inside the models.

        When select_related and related_field exist for the same field being validated, the related
        field is ignored as it won't override the value already collected from the select_related.

        If there is no select_related, then goes through the related field where it **should**
        only return the instance of the the ForeignKey with the ID, making it lazy loaded.

        :return: Model class.
        """
        item: Dict[str, Any] = {}
        select_related = select_related or []
        prefetch_related = prefetch_related or []
        secret_fields = [name for name, field in cls.fields.items() if field.secret] if exclude_secrets else []

        for related in select_related:
            field_name = related.split("__", 1)[0]
            field = cls.meta.fields_mapping[field_name]
            if isinstance(field, RelationshipField):
                model_class, _, remainder = field.traverse_field(related)
            else:
                raise Exception("invalid field")
            if remainder:
                item[field_name] = model_class.from_sqla_row(
                    row,
                    select_related=[remainder],
                    prefetch_related=prefetch_related,
                    exclude_secrets=exclude_secrets,
                    using_schema=using_schema,
                )
            else:
                item[field_name] = model_class.from_sqla_row(row, exclude_secrets=exclude_secrets, using_schema=using_schema)
        # Populate the related names
        # Making sure if the model being queried is not inside a select related
        # This way it is not overritten by any value
        for related, foreign_key in cls.meta.foreign_key_fields.items():
            ignore_related: bool = cls.__should_ignore_related_name(related, select_related)
            if ignore_related or related in secret_fields:
                continue

            columns_to_check = foreign_key.get_column_names(related)
            if secret_fields and not columns_to_check.isdisjoint(secret_fields):
                continue

            model_related = foreign_key.target

            # Apply the schema to the model
            model_related = cls.__apply_schema(model_related, using_schema)

            child_item = {}
            for column_name in columns_to_check:
                if column_name not in row:
                    continue
                elif row[column_name] is not None:  # type: ignore
                    child_item[foreign_key.from_fk_field_name(related, column_name)] = row[column_name]  # type: ignore

            # Make sure we generate a temporary reduced model
            # For the related fields. We simply chnage the structure of the model
            # and rebuild it with the new fields.
            item[related] = model_related.proxy_model(**child_item)

        # Check for the only_fields
        if is_only_fields or is_defer_fields:
            mapping_fields = (
                [str(field) for field in only_fields] if is_only_fields else list(row.keys())  # type: ignore
            )

            for column, value in row._mapping.items():
                if column in secret_fields:
                    continue
                # Making sure when a table is reflected, maps the right fields of the ReflectModel
                if column not in mapping_fields:
                    continue

                if column not in item:
                    item[column] = value

            # We need to generify the model fields to make sure we can populate the
            # model without mandatory fields
            model = cast("Type[Model]", cls.proxy_model(**item))

            # Apply the schema to the model
            model = cls.__apply_schema(model, using_schema)

            model = cls.__handle_prefetch_related(row=row, model=model, prefetch_related=prefetch_related)
            return model
        else:
            # Pull out the regular column values.
            for column in cls.table.columns:
                # Making sure when a table is reflected, maps the right fields of the ReflectModel
                if column.name in secret_fields:
                    continue
                if column.name not in cls.fields.keys():
                    continue
                elif column.name not in item:
                    item[column.name] = row[column]

        model = (
            cast("Type[Model]", cls(**item)) if not exclude_secrets else cast("Type[Model]", cls.proxy_model(**item))
        )

        # Apply the schema to the model
        model = cls.__apply_schema(model, using_schema)

        # Handle prefetch related fields.
        model = cls.__handle_prefetch_related(row=row, model=model, prefetch_related=prefetch_related)
        return model

    @classmethod
    def __apply_schema(cls, model: Type["Model"], schema: Optional[str] = None) -> Type["Model"]:
        # Apply the schema to the model
        if schema is not None:
            model.table = model.build(schema)  # type: ignore
            model.proxy_model.table = model.proxy_model.build(schema)  # type: ignore
        return model

    @classmethod
    def __should_ignore_related_name(cls, related_name: str, select_related: Sequence[str]) -> bool:
        """
        Validates if it should populate the related field if select related is not considered.
        """
        for related_field in select_related:
            fields = related_field.split("__")
            if related_name in fields:
                return True
        return False

    @classmethod
    def __handle_prefetch_related(
        cls,
        row: "Row",
        model: Type["Model"],
        # for advancing
        prefetch_related: Sequence["Prefetch"],
        parent_cls: Optional[Type["Model"]] = None,
        # for going back
        reverse_path: str = "",
        is_nested: bool = False,
    ) -> Type["Model"]:
        """
        Handles any prefetch related scenario from the model.
        Loads in advance all the models needed for a specific record

        Recursively checks for the related field and validates if there is any conflicting
        attribute. If there is, a `QuerySetError` is raised.
        """
        if not parent_cls:
            parent_cls = model

        for related in prefetch_related:
            if not is_nested:
                # Check for conflicting names
                # If to_attr has the same name of any
                if hasattr(parent_cls, related.to_attr):
                    raise QuerySetError(
                        f"Conflicting attribute to_attr='{related.related_name}' with '{related.to_attr}' in {parent_cls.__class__.__name__}"
                    )

            if not is_nested:
                reverse_path = ""

            if "__" in related.related_name:
                first_part, remainder = related.related_name.split("__", 1)

                field = cls.meta.fields_mapping[first_part]
                if isinstance(field, RelationshipField):
                    model_class, reverse_part, remainder = field.traverse_field(related.related_name)
                    if not reverse_part:
                        raise Exception("No reverse relation possible (missing related_name)")
                else:
                    raise Exception("invalid field")

                # Build the new nested Prefetch object
                remainder_prefetch = related.__class__(
                    related_name=remainder, to_attr=related.to_attr, queryset=related.queryset
                )
                if reverse_path:
                    reverse_path = f"{reverse_part}__{reverse_path}"
                else:
                    reverse_path = reverse_part

                # Recursively continue the process of handling the
                # new prefetch
                model_class.__handle_prefetch_related(
                    row,
                    model,
                    prefetch_related=[remainder_prefetch],
                    reverse_path=reverse_path,
                    parent_cls=model,
                    is_nested=True,
                )

            # Check for individual not nested querysets
            elif related.queryset is not None and not is_nested:
                extra = {}
                for pkcol in cls.pkcolumns:
                    filter_by_pk = row[pkcol]
                    extra[f"{related.related_name}__{pkcol}"] = filter_by_pk
                related.queryset.extra = extra

                # Execute the queryset
                records = asyncio.get_event_loop().run_until_complete(cls.run_query(queryset=related.queryset))
                setattr(model, related.to_attr, records)
            else:
                records = cls.process_nested_prefetch_related(
                    row,
                    prefetch_related=related,
                    reverse_path=reverse_path,
                    parent_cls=model,
                    queryset=related.queryset,
                )

                setattr(model, related.to_attr, records)
        return model

    @classmethod
    def process_nested_prefetch_related(
        cls,
        row: "Row",
        prefetch_related: "Prefetch",
        parent_cls: Type["Model"],
        reverse_path: str,
        queryset: "QuerySet",
    ) -> List[Type["Model"]]:
        """
        Processes the nested prefetch related names.
        """
        # Get the related field
        field = cls.meta.fields_mapping[prefetch_related.related_name]
        if isinstance(field, RelationshipField):
            model_class, reverse_part, remainder = field.traverse_field(prefetch_related.related_name)
            if not reverse_part:
                raise Exception("No backward relation possible (missing related_name)")
        else:
            raise Exception("invalid field")

        if reverse_path:
            reverse_path = f"{reverse_part}__{reverse_path}"
        else:
            reverse_path = reverse_part

        # TODO: related_field.clean would be better
        # fix this later when allowing selecting fields for fireign keys
        # Extract foreign key value
        extra = {}
        for pkcol in parent_cls.pkcolumns:
            filter_by_pk = row[pkcol]
            extra[f"{reverse_path}__{pkcol}"] = filter_by_pk

        records = asyncio.get_event_loop().run_until_complete(cls.run_query(model_class, extra, queryset))
        return records

    @classmethod
    async def run_query(
        cls,
        model: Optional[Type["Model"]] = None,
        extra: Optional[Dict[str, Any]] = None,
        queryset: Optional["QuerySet"] = None,
    ) -> Union[List[Type["Model"]], Any]:
        """
        Runs a specific query against a given model with filters.
        """

        if not queryset:
            return await model.query.filter(**extra)  # type: ignore

        if extra:
            queryset.extra = extra

        return await queryset



from functools import cached_property
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

from edgy.core.db.constants import CASCADE
from edgy.core.db.fields.base import BaseField, BaseForeignKey
from edgy.core.db.fields.factories import ForeignKeyFieldFactory
from edgy.core.db.fields.foreign_keys import ForeignKey
from edgy.core.db.relationships.relation import ManyRelation
from edgy.core.terminal import Print
from edgy.core.utils.models import create_edgy_model
from edgy.protocols.many_relationship import ManyRelationProtocol

if TYPE_CHECKING:

    from edgy import Model

T = TypeVar("T", bound="Model")


terminal = Print()


def _removesuffix(text: str, suffix: str) -> str:
    # TODO: replace with _removesuffix when python3.9 is minimum
    if text.endswith(suffix):
        return text[: -len(suffix)]
    else:
        return text


def _removeprefix(text: str, prefix: str) -> str:
    # TODO: replace with removeprefix when python3.9 is minimum
    if text.startswith(prefix):
        return text[len(prefix) :]
    else:
        return text


def _removeprefixes(text:str, *prefixes: Sequence[str]) -> str:
    for prefix in prefixes:
        text = _removeprefix(text, prefix)
    return text

class BaseManyToManyForeignKeyField(BaseForeignKey):
    is_m2m: bool = True
    def __init__(
        self,
        *,
        to_fields: Sequence[str] = (),
        to_foreign_key: str = "",
        from_fields: Sequence[str] = (),
        from_foreign_key: str = "",
        through: Union[str, Type["Model"]] = "",
        embed_through: Union[str, Literal[False]]="",
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.to_fields = to_fields
        self.to_foreign_key = to_foreign_key
        self.from_fields = from_fields
        self.from_foreign_key = from_foreign_key
        self.through = through
        self.embed_through = embed_through

    @cached_property
    def embed_through_prefix(self) -> str:
        if self.embed_through is False:
            return ""
        if not self.embed_through:
            return self.name
        return f"{self.name}__{self.embed_through}"

    @cached_property
    def reverse_embed_through_prefix(self) -> str:
        if self.embed_through is False:
            return ""
        if not self.embed_through:
            return self.reverse_name
        return f"{self.reverse_name}__{self.embed_through}"

    def get_relation(self, **kwargs: Any) -> ManyRelationProtocol:
        return ManyRelation(through=self.through, to=self.to, from_foreign_key=self.from_foreign_key, to_foreign_key=self.to_foreign_key, embed_through=self.embed_through, **kwargs)

    def get_reverse_relation(self, **kwargs: Any) -> ManyRelationProtocol:
        return ManyRelation(through=self.through, to=self.owner, reverse=True, from_foreign_key=self.to_foreign_key, to_foreign_key=self.from_foreign_key, embed_through=self.embed_through, **kwargs)

    def traverse_field(self, path: str) -> Tuple[Any, str, str]:
        if self.embed_through_prefix and path.startswith(self.embed_through_prefix):
            return self.through, self.from_foreign_key, _removeprefixes(path, self.embed_through_prefix, "__")
        return self.to, self.reverse_name, _removeprefixes(path, self.name, "__")

    def reverse_traverse_field_fk(self, path: str) -> Tuple[Any, str, str]:
        # used for target fk
        if self.reverse_embed_through_prefix and path.startswith(self.reverse_embed_through_prefix):
            return self.through, self.to_foreign_key, _removeprefix(_removeprefix(path, self.reverse_embed_through_prefix), "__")
        return self.owner, self.name, _removeprefixes(path, self.reverse_name, "__")

    def add_model_to_register(self, model: Any) -> None:
        """
        Adds the model to the registry to make sure it can be generated by the Migrations
        """
        self.registry.models[model.__name__] = model

    def create_through_model(self) -> None:
        """
        Creates the default empty through model.

        Generates a middle model based on the owner of the field and the field itself and adds
        it to the main registry to make sure it generates the proper models and migrations.
        """
        from edgy.core.db.models.metaclasses import MetaInfo

        self.to = self.target
        __bases__: Tuple[Type["Model"], ...] = ()
        pknames = set()
        if self.through:
            if isinstance(self.through, str):
                self.through = self.registry.models[self.through]
            through = cast(Type["Model"], self.through)
            if through.meta.abstract:
                pknames = set(cast(Sequence[str], through.pknames))
                __bases__ = (through,)
            else:
                if not self.from_foreign_key:
                    candidate = None
                    for field_name, field in through.meta.foreign_key_fields.items():
                        if field.target == self.owner:
                            if candidate:
                                raise ValueError("multiple foreign keys to owner")
                            else:
                                candidate = field_name
                    if not candidate:
                        raise ValueError("no foreign key fo owner found")
                    self.from_foreign_key = candidate
                if not self.to_foreign_key:
                    candidate = None
                    for field_name, field in through.meta.foreign_key_fields.items():
                        if field.target == self.to:
                            if candidate:
                                raise ValueError("multiple foreign keys to target")
                            else:
                                candidate = field_name
                    if not candidate:
                        raise ValueError("no foreign key fo target found")
                    self.to_foreign_key = candidate
                return
        owner_name = self.owner.__name__
        to_name = self.to.__name__
        class_name = f"{owner_name}{to_name}"
        if not self.from_foreign_key:
            self.from_foreign_key = owner_name.lower()

        if not self.to_foreign_key:
            self.to_foreign_key = to_name.lower()

        tablename = f"{owner_name.lower()}s_{to_name}s".lower()
        meta_args = {
            "tablename": tablename,
            "registry": self.registry,
            "multi_related": [to_name.lower()]
        }
        has_pknames = pknames and not pknames.issubset({self.from_foreign_key, self.to_foreign_key})
        if has_pknames:
            meta_args["unique_together"] = [(self.from_foreign_key, self.to_foreign_key)]

        new_meta: MetaInfo = MetaInfo(None, **meta_args)

        to_related_name: Union[str, Literal[False]]
        if self.related_name is False:
            to_related_name = False
        elif self.related_name:
            to_related_name = f"{self.related_name}"
            self.reverse_name = to_related_name
        else:
            if self.unique:
                to_related_name = f"{to_name.lower()}_{class_name.lower()}"
            else:
                to_related_name = f"{to_name.lower()}_{class_name.lower()}s_set"
            self.reverse_name = to_related_name

        # in any way m2m fields will have an index (either by unique_together or by their primary key constraint)

        fields = {
            f"{self.from_foreign_key}": ForeignKey(
                self.owner,
                null=True,
                on_delete=CASCADE,
                related_name=False,
                reverse_name=self.name,
                related_fields=self.from_fields,
                primary_key=not has_pknames,
                index=self.index
            ),
            f"{self.to_foreign_key}": ForeignKey(
                self.to,
                null=True,
                on_delete=CASCADE,
                unique=self.unique,
                related_name=to_related_name,
                related_fields=self.to_fields,
                embed_parent=(self.from_foreign_key, self.embed_through or ""),
                primary_key=not has_pknames,
                index=self.index,
                relation_fn=self.get_reverse_relation,
                reverse_path_fn=self.reverse_traverse_field_fk
            ),
        }

        # Create the through model
        through_model = create_edgy_model(
            __name__=class_name,
            __module__=self.__module__,
            __definitions__=fields,
            __metadata__=new_meta,
            __bases__=__bases__
        )
        self.through = through_model
        self.add_model_to_register(self.through)

    def to_model(self, field_name: str, value: Any, phase: str = "") -> Dict[str, Any]:
        """
        Meta field
        """
        if isinstance(value, ManyRelationProtocol):
            return {field_name: value}
        return {field_name: self.get_relation(refs=value)}

    def has_default(self) -> bool:
        """Checks if the field has a default value set"""
        return False

    def get_default_values(self, field_name: str, cleaned_data: Dict[str, Any]) -> Any:
        """
        Meta field
        """
        return {}

    def __get__(self, instance: "Model", owner: Any = None) -> ManyRelation:
        if instance:
            if self.name not in instance.__dict__:
                instance.__dict__[self.name] = self.get_relation()
            if instance.__dict__[self.name].instance is None:
                instance.__dict__[self.name].instance = instance
            return instance.__dict__[self.name]  # type: ignore
        raise ValueError("Missing instance")

    def __set__(self, instance: "Model", value: Any) -> None:
        relation = self.__get__(instance)
        if not isinstance(value, Sequence):
            value = [value]
        for v in value:
            relation._add_object(v)


class ManyToManyField(ForeignKeyFieldFactory):
    _type: Any = Any
    _bases = (BaseManyToManyForeignKeyField,)

    def __new__(  # type: ignore
        cls,
        to: Union["Model", str],
        *,
        through: Optional["Model"] = None,
        from_fields: Sequence[str] = (),
        to_fields: Sequence[str] = (),
        **kwargs: Any,
    ) -> BaseField:
        for argument in ["null", "on_delete", "on_update"]:
            if kwargs.get(argument, None):
                terminal.write_warning(f"Declaring `{argument}` on a ManyToMany relationship has no effect.")
        kwargs["null"] = True
        kwargs["on_delete"] = CASCADE
        kwargs["on_update"] = CASCADE

        return super().__new__(cls, to=to, through=through, **kwargs)

ManyToMany = ManyToManyField



import copy
from functools import cached_property
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    Iterable,
    List,
    Literal,
    Optional,
    Sequence,
    Set,
    Type,
    Union,
    cast,
)

import sqlalchemy
from pydantic import BaseModel, ConfigDict
from pydantic_core._pydantic_core import SchemaValidator as SchemaValidator
from sqlalchemy.ext.asyncio import AsyncConnection
from typing_extensions import Self

from edgy.core.db.datastructures import Index, UniqueConstraint
from edgy.core.db.models._internal import DescriptiveMeta
from edgy.core.db.models.managers import Manager, RedirectManager
from edgy.core.db.models.metaclasses import BaseModelMeta, MetaInfo
from edgy.core.db.models.model_proxy import ProxyModel
from edgy.core.db.models.utils import build_pkcolumns, build_pknames
from edgy.core.utils.functional import edgy_setattr
from edgy.core.utils.models import DateParser, ModelParser, generify_model_fields
from edgy.core.utils.sync import run_sync
from edgy.exceptions import ImproperlyConfigured

if TYPE_CHECKING:
    from edgy import Model, Registry
    from edgy.core.db.fields.base import BaseField
    from edgy.core.signals import Broadcaster

_empty = cast(Set[str], frozenset())


class EdgyBaseModel(BaseModel, DateParser, ModelParser, metaclass=BaseModelMeta):
    """
    Base of all Edgy models with the core setup.
    """

    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)
    parent: ClassVar[Union[Type[Self], None]]
    is_proxy_model: ClassVar[bool] = False

    query: ClassVar[Manager] = Manager()
    query_related: ClassVar[RedirectManager] = RedirectManager(redirect_name="query")
    meta: ClassVar[MetaInfo] = MetaInfo(None, abstract=True)
    Meta: ClassVar[DescriptiveMeta] = DescriptiveMeta()
    __proxy_model__: ClassVar[Union[Type["Model"], None]] = None
    __db_model__: ClassVar[bool] = False
    __reflected__: ClassVar[bool] = False
    __raw_query__: ClassVar[Optional[str]] = None
    __using_schema__: ClassVar[Union[str, None]] = None
    __show_pk__: ClassVar[bool] = False

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        __show_pk__ = kwargs.pop("__show_pk__", False)
        kwargs = self.transform_input(kwargs, phase="creation")
        super().__init__(**kwargs)
        self.__dict__ = self.setup_model_from_kwargs(kwargs)
        self.__show_pk__ = __show_pk__

    @classmethod
    def transform_input(cls, kwargs: Any, phase: str) -> Any:
        """
        Expand to_model.
        """
        kwargs = {**kwargs}
        new_kwargs: Dict[str, Any] = {}

        fields = cls.meta.fields_mapping
        # phase 1: transform
        for field_name in cls.meta.input_modifying_fields:
            fields[field_name].modify_input(field_name, kwargs)
        # phase 2: apply to_model
        for key, value in kwargs.items():
            field = fields.get(key, None)
            if field is not None:
                new_kwargs.update(**field.to_model(key, value, phase=phase))
            else:
                new_kwargs[key] = value
        return new_kwargs

    def setup_model_from_kwargs(self, kwargs: Any) -> Any:
        """
        Loops and setup the kwargs of the model
        """

        return {
            k: v for k, v in kwargs.items() if k in self.meta.fields_mapping or k in self.meta.model_references
        }

    @property
    def raw_query(self) -> Any:
        return getattr(self, self.__raw_query__)

    @raw_query.setter
    def raw_query(self, value: Any) -> Any:
        edgy_setattr(self, self.raw_query, value)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self}>"

    def __str__(self) -> str:
        pkl = []
        for pkname in self.pknames:
            pkl.append(f"{pkname}={getattr(self, pkname, None)}")
        return f"{self.__class__.__name__}({', '.join(pkl)})"

    @cached_property
    def proxy_model(self) -> Any:
        return self.__class__.proxy_model

    @cached_property
    def identifying_db_fields(self) -> Any:
        """The columns used for loading, can be set per instance defaults to pknames"""
        return self.pkcolumns

    @property
    def can_load(self) -> bool:
        for field in self.identifying_db_fields:
            if self.__dict__.get(field) is None:
                return False
        return True

    @cached_property
    def signals(self) -> "Broadcaster":
        return self.__class__.signals  # type: ignore

    @property
    def table(self) -> sqlalchemy.Table:
        if getattr(self, "_table", None) is None:
            return cast("sqlalchemy.Table", self.__class__.table)
        return self._table

    @table.setter
    def table(self, value: sqlalchemy.Table) -> None:
        try:
            del self._pknames
        except AttributeError:
            pass
        try:
            del self._pkcolumns
        except AttributeError:
            pass
        self._table = value

    @property
    def pkcolumns(self) -> Sequence[str]:
        if self.__dict__.get("_pkcolumns", None) is None:
            if self.__dict__.get("_table", None) is None:
                self._pkcolumns: Sequence[str] = cast(Sequence[str], self.__class__.pkcolumns)
            else:
                build_pkcolumns(self)
        return self._pkcolumns

    @property
    def pknames(self) -> Sequence[str]:
        if self.__dict__.get("_pknames", None) is None:
            if self.__dict__.get("_table", None) is None:
                self._pknames: Sequence[str] = cast(Sequence[str], self.__class__.pknames)
            else:
                build_pknames(self)
        return self._pknames

    def get_columns_for_name(self, name: str) -> Sequence["sqlalchemy.Column"]:
        table = self.table
        meta = self.meta
        if name in meta.field_to_columns:
            return meta.field_to_columns[name]
        elif name in table.columns:
            return (table.columns[name],)
        else:
            return cast(Sequence["sqlalchemy.Column"], _empty)

    def identifying_clauses(self) -> Iterable[Any]:
        for field_name in self.identifying_db_fields:
            field = self.meta.fields_mapping.get(field_name)
            if field is not None:
                for column, value in field.clean(field_name, self.__dict__[field_name]).items():
                    yield getattr(self.table.columns, column) == value
            else:
                yield getattr(self.table.columns, field_name) == self.__dict__[field_name]

    @classmethod
    def generate_proxy_model(cls) -> Type["Model"]:
        """
        Generates a proxy model for each model. This proxy model is a simple
        shallow copy of the original model being generated.
        """
        if cls.__proxy_model__:
            return cls.__proxy_model__

        fields = {key: copy.copy(field) for key, field in cls.meta.fields_mapping.items()}
        proxy_model = ProxyModel(
            name=cls.__name__,
            module=cls.__module__,
            metadata=cls.meta,
            definitions=fields,
        )

        proxy_model.build()
        generify_model_fields(proxy_model.model)
        return proxy_model.model

    def model_dump(self, show_pk: Union[bool, None] = None, **kwargs: Any) -> Dict[str, Any]:
        """
        An updated version of the model dump.
        It can show the pk always and handles the exclude attribute on fields correctly and
        contains the custom logic for fields with getters

        Extra Args:
            show_pk: bool - Enforces showing the primary key in the model_dump.
        """
        # we want a copy
        exclude: Union[Set[str], Dict[str, Any], None] = kwargs.pop("exclude", None)
        if exclude is None:
            initial_full_field_exclude = _empty
            # must be writable
            exclude = set()
        elif isinstance(exclude, dict):
            initial_full_field_exclude = {k for k, v in exclude.items() if v is True}
            exclude = copy.copy(exclude)
        else:
            initial_full_field_exclude = set(exclude)
            exclude = copy.copy(initial_full_field_exclude)

        if isinstance(exclude, dict):
            exclude["__show_pk__"] = True
            for field_name in self.meta.excluded_fields:
                exclude[field_name] = True
        else:
            exclude.update(self.meta.special_getter_fields)
            exclude.update(self.meta.excluded_fields)
            exclude.add("__show_pk__")
        include: Union[Set[str], Dict[str, Any], None] = kwargs.pop("include", None)
        mode: Union[Literal["json", "python"], str] = kwargs.pop("mode", "python")

        should_show_pk = show_pk or self.__show_pk__
        model = dict(super().model_dump(exclude=exclude, include=include, mode=mode, **kwargs))
        # Workaround for metafields, computed field logic introduces many problems
        # so reimplement the logic here
        for field_name in self.meta.special_getter_fields:
            if field_name == "pk":
                continue
            if not should_show_pk or field_name not in self.pknames:
                if field_name in initial_full_field_exclude:
                    continue
                if include is not None and field_name not in include:
                    continue
                if getattr(field_name, "exclude", False):
                    continue
            field: BaseField = self.meta.fields_mapping[field_name]
            try:
                retval = field.__get__(self, self.__class__)
            except AttributeError:
                continue
            sub_include = None
            if isinstance(include, dict):
                sub_include = include.get(field_name, None)
                if sub_include is True:
                    sub_include = None
            sub_exclude = None
            if isinstance(exclude, dict):
                sub_exclude = exclude.get(field_name, None)
                if sub_exclude is True:
                    sub_exclude = None
            if isinstance(retval, BaseModel):
                retval = retval.model_dump(include=sub_include, exclude=sub_exclude, mode=mode, **kwargs)
            else:
                assert (
                    sub_include is None
                ), "sub include filters for CompositeField specified, but no Pydantic model is set"
                assert (
                    sub_exclude is None
                ), "sub exclude filters for CompositeField specified, but no Pydantic model is set"
                if mode == "json" and not getattr(field, "unsafe_json_serialization", False):
                    # skip field if it isn't a BaseModel and the mode is json and unsafe_json_serialization is not set
                    # currently unsafe_json_serialization exists only on ConcreteCompositeFields
                    continue
            alias: str = field_name
            if getattr(field, "serialization_alias", None):
                alias = cast(str, field.serialization_alias)
            elif getattr(field, "alias", None):
                alias = field.alias
            model[alias] = retval
        # tenants cause excluded fields to reappear
        # TODO: find a better bugfix
        for excluded_field in self.meta.excluded_fields:
            model.pop(excluded_field, None)
        return model

    @classmethod
    def build(cls, schema: Optional[str] = None) -> sqlalchemy.Table:
        """
        Builds the SQLAlchemy table representation from the loaded fields.
        """
        tablename: str = cls.meta.tablename  # type: ignore
        registry = cls.meta.registry
        assert registry is not None, "registry is not set"
        metadata: sqlalchemy.MetaData = cast("sqlalchemy.MetaData", registry._metadata)  # type: ignore
        metadata.schema = schema or registry.db_schema

        unique_together = cls.meta.unique_together
        index_constraints = cls.meta.indexes

        columns: List["sqlalchemy.Column"] = []
        global_constraints: List[Any] = []
        for name, field in cls.meta.fields_mapping.items():
            current_columns = field.get_columns(name)
            columns.extend(current_columns)
            global_constraints.extend(field.get_global_constraints(name, current_columns))

        # Handle the uniqueness together
        uniques = []
        for field in unique_together or []:
            unique_constraint = cls._get_unique_constraints(field)
            uniques.append(unique_constraint)

        # Handle the indexes
        indexes = []
        for field in index_constraints or []:
            index = cls._get_indexes(field)
            indexes.append(index)
        return sqlalchemy.Table(
            tablename,
            metadata,
            *columns,
            *uniques,
            *indexes,
            *global_constraints,
            extend_existing=True,
        )

    @classmethod
    def _get_unique_constraints(
        cls, columns: Union[Sequence, str, sqlalchemy.UniqueConstraint]
    ) -> Optional[sqlalchemy.UniqueConstraint]:
        """
        Returns the unique constraints for the model.

        The columns must be a a list, tuple of strings or a UniqueConstraint object.

        :return: Model UniqueConstraint.
        """
        if isinstance(columns, str):
            return sqlalchemy.UniqueConstraint(columns)
        elif isinstance(columns, UniqueConstraint):
            return sqlalchemy.UniqueConstraint(*columns.fields, name=columns.name)
        return sqlalchemy.UniqueConstraint(*columns)

    @classmethod
    def _get_indexes(cls, index: Index) -> Optional[sqlalchemy.Index]:
        """
        Creates the index based on the Index fields
        """
        return sqlalchemy.Index(index.name, *index.fields)  # type: ignore

    def extract_db_fields(self) -> Dict[str, Any]:
        """
        Extracts all the db fields, model references and fields.
        Related fields are not included because they are disjoint.
        """
        fields_mapping = self.meta.fields_mapping
        model_references = self.meta.model_references
        columns = self.__class__.columns
        return {k: v for k, v in self.__dict__.items() if k in fields_mapping or k in columns or k in model_references}

    def get_instance_name(self) -> str:
        """
        Returns the name of the class in lowercase.
        """
        return self.__class__.__name__.lower()

    async def load(self) -> None:
        raise NotImplementedError()

    def __setattr__(self, key: str, value: Any) -> None:
        fields_mapping = self.meta.fields_mapping
        field = fields_mapping.get(key, None)
        if field is not None:
            if hasattr(field, "__set__"):
                # not recommended, better to use to_model instead
                # used in related_fields to mask and not to implement to_model
                field.__set__(self, value)
            else:
                for k, v in field.to_model(key, value, phase="set").items():
                    # bypass __settr__
                    edgy_setattr(self, k, v)
        else:
            # bypass __settr__
            edgy_setattr(self, key, value)

    def __getattr__(self, name: str) -> Any:
        """
        Does following things
        1. Initialize managers on access
        2. Redirects get accesses to getter fields
        3. Run an one off query to populate any foreign key making sure
           it runs only once per foreign key avoiding multiple database calls.
        """
        manager = self.meta.managers.get(name)
        if manager is not None:
            if name not in self.__dict__:
                manager = copy.copy(manager)
                manager.instance = self
                self.__dict__[name] = manager
            return self.__dict__[name]

        field = self.meta.fields_mapping.get(name)
        if field is not None and hasattr(field, "__get__"):
            # no need to set an descriptor object
            return field.__get__(self, self.__class__)
        if name not in self.__dict__ and field is not None and name not in self.identifying_db_fields and self.can_load:
            run_sync(self.load())
            return self.__dict__[name]
        return super().__getattr__(name)

    def __eq__(self, other: Any) -> bool:
        # if self.__class__ != other.__class__:
        #     return False
        # somehow meta gets regenerated, so just compare tablename and registry.
        if self.meta.registry is not other.meta.registry:
            return False
        if self.meta.tablename != other.meta.tablename:
            return False
        self_dict = self._extract_values_from_field(self.extract_db_fields(), is_partial=True)
        other_dict = self._extract_values_from_field(other.extract_db_fields(), is_partial=True)
        key_set = {*self_dict.keys(), *other_dict.keys()}
        for field_name in key_set:
            if self_dict.get(field_name) != other_dict.get(field_name):
                return False
        return True


class EdgyBaseReflectModel(EdgyBaseModel):
    """
    Reflect on async engines is not yet supported, therefore, we need to make a sync_engine
    call.
    """

    __reflected__: ClassVar[bool] = True

    class Meta:
        abstract = True

    @classmethod
    def build(cls, schema: Optional[str] = None) -> Any:
        """
        The inspect is done in an async manner and reflects the objects from the database.
        """
        registry = cls.meta.registry
        assert registry is not None, "registry is not set"
        metadata: sqlalchemy.MetaData = registry._metadata
        schema_name = schema or registry.db_schema
        metadata.schema = schema_name

        tablename: str = cast("str", cls.meta.tablename)
        return run_sync(cls.reflect(registry, tablename, metadata, schema_name))

    @classmethod
    async def reflect(
        cls,
        registry: "Registry",
        tablename: str,
        metadata: sqlalchemy.MetaData,
        schema: Union[str, None] = None,
    ) -> sqlalchemy.Table:
        """
        Reflect a table from the database and return its SQLAlchemy Table object.

        This method connects to the database using the provided registry, reflects
        the table with the given name and metadata, and returns the SQLAlchemy
        Table object.

        Parameters:
            registry (Registry): The registry object containing the database engine.
            tablename (str): The name of the table to reflect.
            metadata (sqlalchemy.MetaData): The SQLAlchemy MetaData object to associate with the reflected table.
            schema (Union[str, None], optional): The schema name where the table is located. Defaults to None.

        Returns:
            sqlalchemy.Table: The reflected SQLAlchemy Table object.

        Raises:
            ImproperlyConfigured: If there is an error during the reflection process.
        """

        def execute_reflection(connection: AsyncConnection) -> sqlalchemy.Table:
            """Helper function to create and reflect the table."""
            try:
                return sqlalchemy.Table(tablename, metadata, schema=schema, autoload_with=connection)
            except Exception as e:
                raise e

        try:
            async with registry.engine.begin() as connection:
                table = await connection.run_sync(execute_reflection)
            await registry.engine.dispose()
            return table
        except Exception as e:
            raise ImproperlyConfigured(detail=str(e)) from e



from typing import TYPE_CHECKING, Any, Dict, Set, Tuple, Type, cast

from pydantic._internal._model_construction import ModelMetaclass

from edgy.conf.module_import import import_string
from edgy.core.db.models.metaclasses import handle_annotations
from edgy.core.marshalls.config import ConfigMarshall
from edgy.core.marshalls.fields import BaseMarshallField
from edgy.core.utils.functional import extract_field_annotations_and_defaults
from edgy.exceptions import MarshallFieldDefinitionError

if TYPE_CHECKING:
    from edgy import Model
    from edgy.core.marshalls import Marshall


class MarshallMeta(ModelMetaclass):
    """
    Base metaclass for the Marshalls
    """

    __slots__ = ()

    def __new__(cls, name: str, bases: Tuple[Type, ...], attrs: Dict[str, Any]) -> Any:
        base_annotations: Dict[str, Any] = {}
        show_pk: bool = False
        marshall_config: ConfigMarshall = attrs.pop("marshall_config", None)
        attrs, model_fields = extract_field_annotations_and_defaults(attrs, BaseMarshallField)

        model_class = super().__new__

        parents = [parent for parent in bases if isinstance(parent, MarshallMeta)]
        if not parents:
            return model_class(cls, name, bases, attrs)

        model_class: "Marshall" = model_class(cls, name, bases, attrs)  # type: ignore
        if name in ("Marshall",):
            return model_class

        if marshall_config is None:
            raise MarshallFieldDefinitionError("The 'marshall_config' was not found. Make sure it is declared and set.")

        # The declared model
        model: "Model" = marshall_config.get("model", None)  # type: ignore
        assert model is not None, "'model' must be declared in the 'ConfigMarshall'."

        if isinstance(cast(str, model), str):
            model: "Model" = import_string(model)  # type: ignore
            marshall_config["model"] = model

        base_fields_include = marshall_config.get("fields", None)
        base_fields_exclude = marshall_config.get("exclude", None)

        assert base_fields_include is None or base_fields_exclude is None, "Use either 'fields' or 'exclude', not both."
        assert (
            base_fields_include is not None or base_fields_exclude is not None
        ), "Either 'fields' or 'exclude' must be declared."

        base_model_fields: Dict[str, Any] = {}

        # Define the fields for the Marshall
        if base_fields_exclude is not None:
            base_model_fields = {k: v for k, v in model.model_fields.items() if k not in base_fields_exclude}
        elif base_fields_include is not None and "__all__" in base_fields_include:
            base_model_fields = {k: v for k, v in model.meta.fields_mapping.items() if k not in model_fields}
            show_pk = True
        else:
            base_model_fields = {
                k: v
                for k, v in model.model_fields.items()
                if k in base_fields_include  # type: ignore
            }

        base_model_fields.update(model_fields)

        # Handles with the fields not declared in the model.
        custom_fields: Dict[str, BaseMarshallField] = {}

        # For custom model_fields
        for k, v in attrs.items():
            if isinstance(v, BaseMarshallField):
                # Make sure the custom fields are flagged.
                if k not in model.meta.fields_mapping:
                    custom_fields[k] = v

        # Handle the check of the custom fields
        for name, field in custom_fields.items():
            if field.__is_method__ and not field.source:
                if not hasattr(model_class, f"get_{name}"):
                    raise MarshallFieldDefinitionError(
                        f"Field '{name}' declared but no 'get_{name}' found in '{model_class.__name__}'."
                    )

        model_class.model_fields = base_model_fields

        # Handle annotations
        annotations: Dict[str, Any] = handle_annotations(bases, base_annotations, attrs)
        model_class.__init_annotations__ = annotations
        model_class.__show_pk__ = show_pk
        model_class.__custom_fields__ = custom_fields
        model_class.marshall_config = marshall_config
        model_class.model_fields.update(base_model_fields)
        model_class.model_rebuild(force=True)

        # Raise for error if any of the required fields is not in the Marshall
        required_fields: Set[str] = {f"'{k}'" for k, v in model.model_fields.items() if not v.null}
        if any(value not in model_class.model_fields.keys() for value in required_fields):
            fields = ", ".join(sorted(required_fields))
            raise MarshallFieldDefinitionError(
                f"'{model.__name__}' model requires the following mandatory fields: [{fields}]."
            )
        return model_class



import click

from edgy.cli.base import stamp as _stamp
from edgy.cli.env import MigrationEnv


@click.option(
    "-d",
    "--directory",
    default=None,
    help=('Migration script directory (default is "migrations")'),
)
@click.option("--sql", is_flag=True, help=("Don't emit SQL to database - dump to standard output " "instead"))
@click.option("--tag", default=None, help=('Arbitrary "tag" name - can be used by custom env.py ' "scripts"))
@click.argument("revision", default="head")
@click.command()
def stamp(env: MigrationEnv, directory: str, sql: bool, tag: str, revision: str) -> None:
    """'stamp' the revision table with the given revision; don't run any
    migrations"""
    _stamp(env.app, directory, revision, sql, tag)



from collections.abc import MutableMapping
from typing import TYPE_CHECKING, Any, Dict, Generator

if TYPE_CHECKING:
    from edgy.core.marshalls import Marshall


class MarshallFieldMapping(MutableMapping):
    """
    Dictionary used to store the information about
    the fields used in the Marshall.
    """

    def __init__(self, marshall: "Marshall") -> None:
        self._marshall = marshall
        self._fields: Dict[str, Any] = {}

    def __setitem__(self, key: str, field: Any) -> None:
        self._fields[key] = field

    def __getitem__(self, key: str) -> Any:
        return self._fields[key]

    def __delitem__(self, key: str) -> None:
        del self._fields[key]

    def __iter__(self) -> Generator:
        return iter(self._fields)

    def __len__(self) -> int:
        return len(self._fields)

    def __repr__(self) -> Any:
        return dict.__repr__(self._fields)



from contextvars import ContextVar
from typing import TYPE_CHECKING, Type, Union

if TYPE_CHECKING:
    from edgy import Database, Model, QuerySet

TENANT: ContextVar[str] = ContextVar("tenant", default=None)
SHEMA: ContextVar[str] = ContextVar("SHEMA", default=None)


def get_tenant() -> Union[str, None]:
    """
    Gets the current active tenant in the context.
    """
    return TENANT.get()


def set_tenant(value: Union[str, None]) -> None:
    """
    Sets the global tenant for the context of the queries.
    When a global tenant is set the `get_context_schema` -> `SCHEMA` is ignored.
    """
    TENANT.set(value)


def get_schema() -> Union[str, None]:
    return SHEMA.get()


def set_schema(value: Union[str, None]) -> None:
    SHEMA.set(value)


def set_queryset_schema(
    queryset: "QuerySet",
    model_class: Type["Model"],
    value: Union[str, None],
) -> "QuerySet":
    """
    Returns a new queryset object pointing to the desired schema of the
    using.
    """
    return queryset.__class__(
        model_class=model_class,
        using_schema=value,
        table=model_class.table_schema(value),
    )


def set_queryset_database(
    queryset: "QuerySet",
    model_class: Type["Model"],
    database: Type["Database"],
    schema: Union[str, None] = None,
) -> "QuerySet":
    """
    Returns a new queryset object pointing to the desired schema of the
    using.
    """
    if not schema:
        return queryset.__class__(
            model_class=model_class,
            database=database,
            table=model_class.table_schema(schema),
        )
    return queryset.__class__(model_class=model_class, database=database)



from typing import TYPE_CHECKING, Any, Optional, Protocol, runtime_checkable

if TYPE_CHECKING:  # pragma: nocover
    from edgy import Model


@runtime_checkable
class ManyRelationProtocol(Protocol):
    instance: Any

    """Defines the what needs to be implemented when using the ManyRelationProtocol"""
    async def save_related(self) -> None: ...

    async def add(self, child:  Optional["Model"]=None) -> Optional["Model"]: ...

    async def remove(self, child: Optional["Model"]=None) -> None: ...



from typing import TYPE_CHECKING, Any, Iterable, Set, cast

from edgy.core.connection.registry import Registry

if TYPE_CHECKING:
    from edgy.core.db.models.base import EdgyBaseModel
    from edgy.core.db.models.model import Model


def get_model(registry: Registry, model_name: str) -> "Model":
    """
    Return the model with capitalize model_name.

    Raise lookup error if no model is found.
    """
    try:
        return cast("Model", registry.models[model_name])
    except KeyError:
        raise LookupError(f"Registry doesn't have a {model_name} model.") from None


def build_pknames(model_class: Any) -> None:
    """
    Set explicit pknames

    Raise error if primary key column has no field associated.
    """
    meta = model_class.meta
    pknames: Set[str] = set()
    for field_name, field in meta.fields_mapping.items():
        if field.primary_key:
            pknames.add(field_name)
    model_class._pknames = tuple(sorted(pknames))


def build_pkcolumns(model_class: Any) -> None:
    """
    Set pkcolumns

    Raise error if primary key column has no field associated.
    """
    table = model_class.table
    pkcolumns: Set[str] = set()
    for column in table.columns:
        if column.primary_key:
            # key is the sqlalchemy name, in our case name and key should be identically
            pkcolumns.add(column.key)
    model_class._pkcolumns = tuple(sorted(pkcolumns))


def from_model_to_clauses(model: "EdgyBaseModel") -> Iterable[Any]:
    for column in model.columns_for_load:
        yield getattr(model.table.columns, column) == getattr(model, column)



from functools import cached_property
from typing import Dict, List, Set

from pydantic_settings import BaseSettings, SettingsConfigDict


class EdgySettings(BaseSettings):
    model_config = SettingsConfigDict(extra="allow", ignored_types=(cached_property,))
    ipython_args: List[str] = ["--no-banner"]
    ptpython_config_file: str = "~/.config/ptpython/config.py"

    # Dialects
    postgres_dialects: Set[str] = {"postgres", "postgresql"}
    mysql_dialects: Set[str] = {"mysql"}
    sqlite_dialects: Set[str] = {"sqlite"}
    mssql_dialects: Set[str] = {"mssql"}

    # Drivers
    postgres_drivers: Set[str] = {"aiopg", "asyncpg"}
    mysql_drivers: Set[str] = {"aiomysql", "asyncmy"}
    sqlite_drivers: Set[str] = {"aiosqlite"}

    @property
    def mssql_drivers(self) -> Set[str]:
        """
        Do not override this one as SQLAlchemy doesn't support async for MSSQL.
        """
        return {"aioodbc"}

    # General settings
    filter_operators: Dict[str, str] = {
        "exact": "__eq__",
        "iexact": "ilike",
        "contains": "like",
        "icontains": "ilike",
        "in": "in_",
        "gt": "__gt__",
        "gte": "__ge__",
        "lt": "__lt__",
        "lte": "__le__",
    }
    many_to_many_relation: str = "relation_{key}"
    dialects: Dict[str, str] = {"postgres": "postgres", "postgresql": "postgresql"}



from databasez import Database as Databasez  # noqa
from databasez import DatabaseURL as DatabaseURL  # noqa


class Database(Databasez):
    """
    An abstraction on the top of the EncodeORM databases.Database object.

    This object allows to pass also a configuration dictionary in the format of

    ```python
    DATABASEZ_CONFIG = {
        "connection": {
            "credentials": {
                "scheme": 'sqlite', "postgres"...
                "host": ...,
                "port": ...,
                "user": ...,
                "password": ...,
                "database": ...,
                "options": {
                    "driver": ...
                    "ssl": ...
                }
            }
        }
    }
    ```
    """

    ...



from abc import ABC, abstractmethod
from typing import Any


class BaseExtra(ABC):
    @abstractmethod
    def set_edgy_extension(self, app: Any, **kwargs: Any) -> None:
        raise NotImplementedError("Any class implementing the extra must implement set_edgy_extension() .")



import typing
from datetime import datetime
from enum import Enum
from typing import TYPE_CHECKING, Any, Dict, Optional, Tuple, Type

from orjson import OPT_OMIT_MICROSECONDS, OPT_SERIALIZE_NUMPY, dumps
from pydantic import ConfigDict

import edgy
from edgy.core.db.fields.base import BaseField

if TYPE_CHECKING:
    from edgy import Model
    from edgy.core.db.models.metaclasses import MetaInfo

type_ignored_setattr = setattr

def _has_auto_now(field: Type[BaseField]) -> bool:
    """
    Checks if the field is auto now
    """
    return True if hasattr(field, "auto_now") and field.auto_now else False


def _has_auto_now_add(field: Type[BaseField]) -> bool:
    """
    Checks if the field is auto now add
    """
    return True if hasattr(field, "auto_now_add") and field.auto_now_add else False


def _is_datetime(field: Type[BaseField]) -> bool:
    """
    Validates if the field type is a datetime type.
    """
    return bool(field.field_type == datetime)


class DateParser:
    def _update_auto_now_fields(self, values: Any, fields: Any) -> Any:
        """
        Updates the `auto_now` fields
        """
        for name, field in fields.items():
            if isinstance(field, BaseField) and _has_auto_now(field) and _is_datetime(field):
                values.update(field.get_default_values(name, values))
        return values

    def _resolve_value(self, value: typing.Any) -> typing.Any:
        if isinstance(value, dict):
            return dumps(
                value,
                option=OPT_SERIALIZE_NUMPY | OPT_OMIT_MICROSECONDS,
            ).decode("utf-8")
        elif isinstance(value, Enum):
            return value.name
        return value


class ModelParser:
    def _extract_model_references(self, extracted_values: Any, model_class: Optional[Type["Model"]]) -> Any:
        """
        Exracts any possible model references from the EdgyModel and returns a dictionary.
        """
        model_references = {
            name: extracted_values.get(name, None)
            for name in model_class.meta.model_references.keys()  # type: ignore
            if extracted_values.get(name)
        }
        return model_references

    def _extract_values_from_field(
        self,
        extracted_values: Any,
        model_class: Optional["Model"] = None,
        is_update: bool = False,
        is_partial: bool = False,
    ) -> Any:
        """
        Extracts all the default values from the given fields and returns the raw
        value corresponding to each field.
        """
        model_cls = model_class or self
        validated: Dict[str, Any] = {}
        # phase 1: transform when required
        if model_cls.meta.input_modifying_fields:
            extracted_values = {**extracted_values}
            for field_name in model_cls.meta.input_modifying_fields:
                model_cls.fields[field_name].modify_input(field_name, extracted_values)
        # phase 2: validate fields and set defaults for readonly
        for field_name, field in model_cls.fields.items():  # type: ignore
            if not is_partial and field.read_only:
                if field.has_default():
                    if not is_update:
                        validated.update(field.get_default_values(field_name, validated))
                    else:
                        # For datetimes with `auto_now` and `auto_now_add`
                        if not _has_auto_now_add(field):
                            validated.update(field.get_default_values(field_name, validated))
                continue
            if field_name in extracted_values:
                item = extracted_values[field_name]
                for sub_name, value in field.clean(field_name, item).items():
                    if sub_name in validated:
                        raise ValueError(f"value set twice for key: {sub_name}")
                    validated[sub_name] = value

        # phase 3: set defaults for the rest if not an update
        if not is_partial:
            for field_name, field in model_cls.fields.items():  # type: ignore
                # we need a second run
                if not field.read_only and field_name not in validated:
                    if field.has_default():
                        validated.update(field.get_default_values(field_name, validated))
        # Update with any ModelRef
        validated.update(self._extract_model_references(extracted_values, model_cls))
        return validated


def create_edgy_model(
    __name__: str,
    __module__: str,
    __definitions__: Optional[Dict[Any, Any]] = None,
    __metadata__: Optional[Type["MetaInfo"]] = None,
    __qualname__: Optional[str] = None,
    __config__: Optional[ConfigDict] = None,
    __bases__: Optional[Tuple[Type["Model"]]] = None,
    __proxy__: bool = False,
    __pydantic_extra__: Any = None,
) -> Type["Model"]:
    """
    Generates an `edgy.Model` with all the required definitions to generate the pydantic
    like model.
    """

    if not __bases__:
        __bases__ = (edgy.Model,)

    qualname = __qualname__ or __name__
    core_definitions = {
        "__module__": __module__,
        "__qualname__": qualname,
        "is_proxy_model": __proxy__,
    }
    if not __definitions__:
        __definitions__ = {}

    core_definitions.update(**__definitions__)

    if __config__:
        core_definitions.update(**{"model_config": __config__})
    if __metadata__:
        core_definitions.update(**{"Meta": __metadata__})
    if __pydantic_extra__:
        core_definitions.update(**{"__pydantic_extra__": __pydantic_extra__})

    model: Type["Model"] = type(__name__, __bases__, core_definitions)
    return model


def generify_model_fields(model: Type["Model"]) -> Dict[Any, Any]:
    """
    Makes all fields generic when a partial model is generated or used.
    This also removes any metadata for the field such as validations making
    it a clean slate to be used internally to process dynamic data and removing
    the constraints of the original model fields.
    """
    fields = {}

    # handle the nested non existing results
    for name, field in model.model_fields.items():
        type_ignored_setattr(field, "annotation", Any)
        type_ignored_setattr(field, "null", True)
        type_ignored_setattr(field, "metadata", [])
        fields[name] = field
    return fields



import uuid
from datetime import date
from typing import Any, Dict, Union, cast
from uuid import UUID

from loguru import logger

import edgy
from edgy import settings
from edgy.contrib.multi_tenancy.utils import create_tables
from edgy.core.db.models.model import Model
from edgy.core.db.models.utils import get_model
from edgy.exceptions import ModelSchemaError, ObjectNotFound


class TenantMixin(edgy.Model):
    """
    Abstract table that acts as an entry-point for
    the tenants with Edgy contrib.
    """

    schema_name: str = edgy.CharField(max_length=63, unique=True, index=True)
    domain_url: str = edgy.URLField(null=True, default=settings.domain, max_length=2048)
    tenant_name: str = edgy.CharField(max_length=100, unique=True, null=False)
    tenant_uuid: UUID = edgy.UUIDField(default=uuid.uuid4, null=False)
    paid_until: date = edgy.DateField(null=True)
    on_trial: bool = edgy.BooleanField(null=True)  # type: ignore
    created_on: date = edgy.DateField(auto_now_add=True)

    # Default True, the schema will be automatically created and synched when it is saved.
    auto_create_schema: bool = getattr(settings, "auto_create_schema", True)
    """
    Set this flag to false on a parent class if you don't want the schema to be automatically
    generated.
    """
    auto_drop_schema: bool = getattr(settings, "auto_drop_schema", False)
    """
    Use with caution! Set this flag to true if you want the schema to be dropped if the
    tenant row is deleted from this table.
    """

    class Meta:
        abstract = True

    def __str__(self) -> str:
        return f"{self.tenant_name} - {self.schema_name}"

    async def save(
        self: Any,
        force_save: bool = False,
        values: Dict[str, Any] = None,
        **kwargs: Any,
    ) -> Model:
        """
        Creates a tenant record and generates a schema in the database.

        When a schema is created, then generates the tables for that same schema
        from the tenant models.
        """
        fields = self.extract_db_fields()
        schema_name = fields.get("schema_name", None)

        if (
            not schema_name
            or schema_name.lower() == settings.tenant_schema_default.lower()
            or schema_name == self.meta.registry.db_schema
        ):
            current_schema = (
                settings.tenant_schema_default.lower()
                if not self.meta.registry.db_schema
                else self.meta.registry.db_schema
            )
            raise ModelSchemaError(
                "Can't update tenant outside it's own schema or the public schema. Current schema is '%s'"
                % current_schema
            )

        tenant = await super().save(force_save, values, **kwargs)
        registry = self.meta.registry
        assert registry is not None, "registry is not set"
        try:
            await registry.schema.create_schema(schema=tenant.schema_name, if_not_exists=True)
            await create_tables(registry, registry.tenant_models, tenant.schema_name)
        except Exception as e:
            message = f"Rolling back... {str(e)}"
            logger.error(message)
            await self.delete()
        return cast(edgy.Model, tenant)

    async def delete(self, force_drop: bool = False) -> None:
        """
        Validates the permissions for the schema before deleting it.
        """
        if self.schema_name == settings.tenant_schema_default:
            raise ValueError("Cannot drop public schema.")
        registry = self.meta.registry
        assert registry is not None, "registry is not set"

        await registry.schema.drop_schema(schema=self.schema_name, cascade=True, if_exists=True)
        await super().delete()


class DomainMixin(edgy.Model):
    """
    All models that store the domains must use this class
    """

    domain = edgy.CharField(max_length=253, unique=True, db_index=True)
    tenant = edgy.ForeignKey(settings.tenant_model, index=True, related_name="domains")
    is_primary = edgy.BooleanField(default=True, index=True)

    class Meta:
        abstract = True

    def __str__(self) -> str:
        return self.domain

    async def save(
        self: Any,
        force_save: bool = False,
        values: Dict[str, Any] = None,
        **kwargs: Any,
    ) -> Model:
        async with self.meta.registry.database.transaction():
            domains = self.__class__.query.filter(tenant=self.tenant, is_primary=True).exclude(pk=self.pk)

            exists = await domains.exists()

            self.is_primary = self.is_primary or (not exists)
            if self.is_primary:
                await domains.update(is_primary=False)

            return await super().save(force_save, values, **kwargs)

    async def delete(self) -> None:
        tenant = await self.tenant.load()
        if tenant.schema_name.lower() == settings.tenant_schema_default.lower() and self.domain == settings.domain_name:
            raise ValueError("Cannot drop public domain.")
        return await super().delete()


class TenantUserMixin(edgy.Model):
    """
    Mapping between user and a client (tenant).
    """

    user = edgy.ForeignKey(
        settings.auth_user_model,
        null=False,
        blank=False,
        related_name="tenant_user_users",
    )
    tenant = edgy.ForeignKey(
        settings.tenant_model,
        null=False,
        blank=False,
        related_name="tenant_users_tenant",
    )
    is_active = edgy.BooleanField(default=False)
    created_on = edgy.DateField(auto_now_add=True)

    class Meta:
        abstract = True

    @classmethod
    async def get_active_user_tenant(cls, user: edgy.Model) -> Union[edgy.Model, None]:
        """
        Obtains the active user tenant.
        """
        try:
            tenant = await get_model(
                registry=cls.meta.registry, model_name=cls.__name__
            ).query.get(user=user, is_active=True)
            await tenant.tenant.load()

        except ObjectNotFound:
            return None
        return cast(edgy.Model, tenant.tenant)

    def __str__(self) -> str:
        return f"User: {self.user.pk}, Tenant: {self.tenant}"

    async def save(self, *args: Any, **kwargs: Any) -> edgy.Model:
        await super().save(*args, **kwargs)
        if self.is_active:
            await (
                get_model(
                    registry=self.meta.registry, model_name=self.__class__.__name__
                )
                .query.filter(is_active=True, user=self.user)
                .exclude(pk=self.pk)
                .update(is_active=False)
            )
        return self



import typing


class EdgyException(Exception):
    def __init__(
        self,
        *args: typing.Any,
        detail: str = "",
    ):
        self.detail = detail
        super().__init__(*(str(arg) for arg in args if arg), self.detail)

    def __repr__(self) -> str:
        if self.detail:
            return f"{self.__class__.__name__} - {self.detail}"
        return self.__class__.__name__

    def __str__(self) -> str:
        return "".join(self.args).strip()


class FieldDefinitionError(EdgyException): ...


class MarshallFieldDefinitionError(FieldDefinitionError): ...


class ObjectNotFound(EdgyException): ...


class MultipleObjectsReturned(EdgyException): ...


class ImproperlyConfigured(EdgyException): ...


class ForeignKeyBadConfigured(EdgyException): ...


class RelationshipIncompatible(EdgyException): ...


class DuplicateRecordError(EdgyException): ...


class RelationshipNotFound(EdgyException): ...


class QuerySetError(EdgyException): ...


class ModelReferenceError(EdgyException): ...


class SchemaError(EdgyException): ...


class SignalError(EdgyException): ...


class CommandEnvironmentError(EdgyException): ...


class ModelSchemaError(EdgyException): ...



import copy
import inspect
from collections import UserDict, deque
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    FrozenSet,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
    cast,
    get_origin,
)

import sqlalchemy
from pydantic._internal._model_construction import ModelMetaclass

from edgy.core import signals as signals_module
from edgy.core.connection.registry import Registry
from edgy.core.db import fields as edgy_fields
from edgy.core.db.datastructures import Index, UniqueConstraint
from edgy.core.db.fields.base import BaseField, PKField
from edgy.core.db.fields.foreign_keys import BaseForeignKeyField
from edgy.core.db.fields.many_to_many import BaseManyToManyForeignKeyField
from edgy.core.db.fields.ref_foreign_key import BaseRefForeignKeyField
from edgy.core.db.models.managers import BaseManager
from edgy.core.db.models.utils import build_pkcolumns, build_pknames
from edgy.core.db.relationships.related_field import RelatedField
from edgy.core.utils.functional import extract_field_annotations_and_defaults
from edgy.exceptions import ForeignKeyBadConfigured, ImproperlyConfigured

if TYPE_CHECKING:
    from edgy.core.db.models import Model, ModelRef, ReflectModel

_empty_dict: Dict[str, Any] = {}
_empty_set: FrozenSet[Any] = frozenset()

class FieldToColumns(UserDict, Dict[str, Sequence["sqlalchemy.Column"]]):
    def __init__(self, meta: "MetaInfo"):
        self.meta = meta
        super().__init__()

    def __getitem__(self, name: str) -> Sequence["sqlalchemy.Column"]:
        if name in self.data:
            return cast(Sequence["sqlalchemy.Column"], self.data[name])
        field = self.meta.fields_mapping[name]
        result = self.data[name] = field.get_columns(name)
        return result

    def __setitem__(self, name: str, value: Any) -> None:
        raise Exception("Cannot set item here")

    def __iter__(self) -> Any:
        self.meta.columns_to_field.init()
        return super().__iter__()

    def __contains__(self, key: str) -> bool:
        try:
            self[key]
            return True
        except KeyError:
            return False


class FieldToColumnNames(FieldToColumns, Dict[str, FrozenSet[str]]):
    def __getitem__(self, name: str) -> FrozenSet[str]:
        if name in self.data:
            return cast(FrozenSet[str], self.data[name])
        column_names = frozenset(column.name for column in self.meta.field_to_columns[name])
        result = self.data[name] = column_names
        return result


class ColumnsToField(UserDict, Dict[str, str]):
    def __init__(self, meta: "MetaInfo"):
        self.meta = meta
        self._init = False
        super().__init__()

    def init(self) -> None:
        if not self._init:
            self._init = True
            _columns_to_field: Dict[str, str] = {}
            for field_name in self.meta.fields_mapping.keys():
                # init structure
                column_names = self.meta.field_to_column_names[field_name]
                for column_name in column_names:
                    if column_name in _columns_to_field:
                        raise ValueError(f"column collision: {column_name} between field {field_name} and {_columns_to_field[column_name]}")
                    _columns_to_field[column_name] = field_name
            self.data.update(_columns_to_field)

    def __getitem__(self, name: str) -> str:
        self.init()
        return cast(str, super().__getitem__(name))

    def __setitem__(self, name: str, value: Any) -> None:
        raise Exception("Cannot set item here")

    def __contains__(self, name: str) -> bool:
        self.init()
        return super().__contains__(name)


_trigger_attributes_MetaInfo = {
    "field_to_columns",
    "field_to_column_names",
    "foreign_key_fields",
    "columns_to_field",
    "special_getter_fields",
    "input_modifying_fields",
    "excluded_fields",
}

class MetaInfo:
    __slots__ = (
        "abstract",
        "inherit",
        "fields_mapping",
        "registry",
        "tablename",
        "unique_together",
        "indexes",
        "parents",
        "model",
        "managers",
        "multi_related",
        "model_references",
        "signals",
        "input_modifying_fields",
        "foreign_key_fields",
        "field_to_columns",
        "field_to_column_names",
        "columns_to_field",
        "special_getter_fields",
        "excluded_fields",
        "_is_init"
    )
    _include_dump = (*filter(lambda x: x not in {
        "field_to_columns",
        "field_to_column_names",
        "columns_to_field",
        "_is_init"
    }, __slots__), "pk", "is_multi")

    field_to_columns: FieldToColumns
    field_to_column_names: FieldToColumnNames
    columns_to_field: Dict[str, str]

    def __init__(self, meta: Any = None, **kwargs: Any) -> None:
        #  Difference between meta extraction and kwargs: meta attributes are copied
        self.abstract: bool = getattr(meta, "abstract", False)
        # for embedding
        self.inherit: bool = getattr(meta, "inherit", True)
        self.registry: Optional[Registry] = getattr(meta, "registry", None)
        self.tablename: Optional[str] = getattr(meta, "tablename", None)
        self.unique_together: Any = getattr(meta, "unique_together", None)
        self.indexes: Any = getattr(meta, "indexes", None)
        self.signals = signals_module.Broadcaster(getattr(meta, "signals", None) or {})
        self.signals.set_lifecycle_signals_from(signals_module, overwrite=False)
        self.parents: List[Any] = [*getattr(meta, "parents", _empty_set)]
        self.fields_mapping: Dict[str, BaseField] = {**getattr(meta, "fields_mapping", _empty_dict)}
        self.model_references: Dict[str, "ModelRef"] = {**getattr(meta, "model_references", _empty_dict)}
        self.managers: Dict[str, BaseManager] = {**getattr(meta, "managers", _empty_dict)}
        self.multi_related: List[str] =  [*getattr(meta, "multi_related", _empty_set)]
        self.model: Optional[Type["Model"]] = None
        self.load_dict(kwargs)

    @property
    def pk(self) -> Optional[PKField]:
        return cast(Optional[PKField], self.fields_mapping.get("pk"))

    @property
    def is_multi(self) -> bool:
        return bool(self.multi_related)

    def model_dump(self) -> Dict[Any, Any]:
        return {k: getattr(self, k, None) for k in self._include_dump}

    def load_dict(self, values: Dict[str, Any], _init: bool=False) -> None:
        """
        Loads the metadata from a dictionary.
        """
        for key, value in values.items():
            # we want triggering invalidate in case it is fields_mapping
            setattr(self, key, value)

    def __setattr__(self, name: str, value: Any) -> None:
        super().__setattr__(name, value)
        if name == "fields_mapping":
            self.invalidate()

    def __getattribute__(self, name: str) -> Any:
        # lazy execute
        if name in _trigger_attributes_MetaInfo and not self._is_init:
            self.init_fields_mapping()
        return super().__getattribute__(name)

    def init_fields_mapping(self) -> None:
        special_getter_fields = set()
        excluded_fields = set()
        input_modifying_fields = set()
        foreign_key_fields: Dict[str, BaseField] = {}
        for key, field in self.fields_mapping.items():
            if hasattr(field, "__get__"):
                special_getter_fields.add(key)
            if getattr(field, "exclude", False):
                excluded_fields.add(key)
            if hasattr(field, "modify_input"):
                input_modifying_fields.add(key)
            if isinstance(field, BaseForeignKeyField):
                foreign_key_fields[key] = field
        self.special_getter_fields: FrozenSet[str] = frozenset(special_getter_fields)
        self.excluded_fields: FrozenSet[str] = frozenset(excluded_fields)
        self.input_modifying_fields: FrozenSet[str] = frozenset(input_modifying_fields)
        self.foreign_key_fields: Dict[str, BaseField] = foreign_key_fields
        self.field_to_columns = FieldToColumns(self)
        self.field_to_column_names = FieldToColumnNames(self)
        self.columns_to_field = ColumnsToField(self)
        self._is_init = True

    def invalidate(self, clear_class_attrs: bool=True) -> None:
        self._is_init = False
        # prevent cycles and mem-leaks
        self.field_to_columns = FieldToColumns(self)
        self.field_to_column_names = FieldToColumnNames(self)
        self.columns_to_field = ColumnsToField(self)
        if clear_class_attrs:
            for attr in ("_table", "_pknames", "_pkcolumns"):
                try:
                    delattr(self.model, attr)
                except AttributeError:
                    pass

    def full_init(self, init_column_mappers: bool=True, init_class_attrs: bool=True) -> None:
        if not self._is_init:
            self.init_fields_mapping()
        if init_column_mappers:
            self.columns_to_field.init()
        if init_class_attrs:
            for attr in ("table", "pknames", "pkcolumns"):
                getattr(self.model, attr)

    def get_columns_for_name(self, name: str) -> Sequence["sqlalchemy.Column"]:
        if name in self.field_to_columns:
            return self.field_to_columns[name]
        elif self.model and name in self.model.table.columns:
            return (self.model.table.columns[name],)
        else:
            return cast(Sequence["sqlalchemy.Column"], _empty_set)


def get_model_registry(
    bases: Tuple[Type, ...], meta_class: Optional[Union["object", MetaInfo]] = None
) -> Optional[Registry]:
    """
    When a registry is missing from the Meta class, it should look up for the bases
    and obtain the first found registry.
    """
    if meta_class is not None:
        direct_registry: Optional[Registry] = getattr(meta_class, "registry", None)
        if direct_registry is not None:
            return direct_registry

    for base in bases:
        meta: MetaInfo = getattr(base, "meta", None)  # type: ignore
        if not meta:
            continue
        found_registry: Optional[Registry] = getattr(meta, "registry", None)

        if found_registry is not None:
            return found_registry
    return None


def _set_related_name_for_foreign_keys(
    foreign_keys: Dict[str, BaseForeignKeyField],
    model_class: Union["Model", "ReflectModel"],
) -> None:
    """
    Sets the related name for the foreign keys.
    When a `related_name` is generated, creates a RelatedField from the table pointed
    from the ForeignKey declaration and the the table declaring it.
    """
    if not foreign_keys:
        return

    for name, foreign_key in foreign_keys.items():
        related_name = getattr(foreign_key, "related_name", None)
        if related_name is False:
            # skip related_field
            continue

        if not related_name:
            if foreign_key.unique:
                related_name = f"{model_class.__name__.lower()}"
            else:
                related_name = f"{model_class.__name__.lower()}s_set"

        if related_name in foreign_key.target.meta.fields_mapping:
            raise ForeignKeyBadConfigured(
                f"Multiple related_name with the same value '{related_name}' found to the same target. Related names must be different."
            )
        foreign_key.related_name = related_name
        foreign_key.reverse_name = related_name

        related_field = RelatedField(
            foreign_key_name=name,
            name=related_name,
            owner=foreign_key.target,
            related_from=model_class,
        )

        # Set the related name
        target = foreign_key.target
        target.meta.fields_mapping[related_name] = related_field


def _handle_annotations(base: Type, base_annotations: Dict[str, Any]) -> None:
    for parent in base.__mro__[1:]:
        _handle_annotations(parent, base_annotations)
    if hasattr(base, "__init_annotations__") and base.__init_annotations__:
        base_annotations.update(base.__init_annotations__)
    elif hasattr(base, "__annotations__") and base.__annotations__:
        base_annotations.update(base.__annotations__)


def handle_annotations(
    bases: Tuple[Type, ...], base_annotations: Dict[str, Any], attrs: Any
) -> Dict[str, Any]:
    """
    Handles and copies some of the annotations for
    initialiasation.
    """
    for base in bases:
        _handle_annotations(base, base_annotations)

    annotations: Dict[str, Any] = (
        copy.copy(attrs["__init_annotations__"])
        if "__init_annotations__" in attrs
        else copy.copy(attrs["__annotations__"])
    )
    annotations.update(base_annotations)
    return annotations


_occluded_sentinel = object()

def _extract_fields_and_managers(base: Type, attrs: Dict[str, Any]) -> None:
    from edgy.core.db.fields.composite_field import CompositeField
    meta: Union[MetaInfo, None] = getattr(base, "meta", None)
    if not meta:
        # Mixins and other classes
        # Note: from mixins BaseFields and BaseManagers are imported despite inherit=False until a model in the
        # hierarchy uses them
        # Here is _occluded_sentinel not overwritten
        for key, value in inspect.getmembers(base):
            if key not in attrs:
                if isinstance(value, BaseField):
                    attrs[key] = value
                elif isinstance(value, BaseManager):
                    attrs[key] = value.__class__()
                elif isinstance(value, BaseModelMeta):
                    attrs[key] = CompositeField(inner_fields=value, prefix_embedded=f"{key}_", inherit=value.meta.inherit, name=key, owner=value)
            elif attrs[key] is _occluded_sentinel:
                # when occluded only include if inherit is True
                if isinstance(value, BaseField) and value.inherit:
                    attrs[key] = value
                elif isinstance(value, BaseManager) and value.inherit:
                    attrs[key] = value.__class__()
                elif isinstance(value, BaseModelMeta) and value.meta.inherit:
                    attrs[key] = CompositeField(inner_fields=value, prefix_embedded=f"{key}_", inherit=value.meta.inherit, name=key, owner=value)

    else:
        # abstract classes
        for key, value in meta.fields_mapping.items():
            if key not in attrs:
                # when abstract or inherit passthrough
                if meta.abstract or value.inherit:
                    attrs[key] = value
                    assert value.owner is not None
                else:
                    attrs[key] = _occluded_sentinel
            elif attrs[key] is _occluded_sentinel and value.inherit:
                # when occluded only include if inherit is True
                attrs[key] = value
                assert value.owner is not None
        for key, value in meta.managers.items():
            if key not in attrs:
                # when abstract or inherit passthrough
                if meta.abstract or value.inherit:
                    attrs[key] = value
                else:
                    attrs[key] = _occluded_sentinel
            elif attrs[key] is _occluded_sentinel and value.inherit:
                # when occluded only include if inherit is True
                attrs[key] = value

    # from strongest to weakest
    for parent in base.__mro__[1:]:
        _extract_fields_and_managers(parent, attrs)


def extract_fields_and_managers(bases: Sequence[Type], attrs: Optional[Dict[str, Any]] = None) -> Dict[str, Any]:
    """
    Search for fields and managers and return them.

    Managers are copied.

    Note: managers and fields with inherit=False are still extracted from mixins as long there is no intermediate model
    """

    from edgy.core.db.fields.composite_field import CompositeField
    attrs = {} if attrs is None else {**attrs}
    # order is important
    for key in list(attrs.keys()):
        value = attrs[key]
        if isinstance(value, BaseModelMeta):
            value = attrs[key]
            attrs[key] = CompositeField(inner_fields=value, prefix_embedded=f"{key}_", inherit=value.meta.inherit, owner=value)
    for base in bases:
        _extract_fields_and_managers(base, attrs)
    # now remove sentinels
    for key in list(attrs.keys()):
        value = attrs[key]
        if value is _occluded_sentinel:
            attrs.pop(key)
    return attrs


class BaseModelMeta(ModelMetaclass):
    __slots__ = ()

    def __new__(cls, name: str, bases: Tuple[Type, ...], attrs: Dict[str, Any]) -> Any:
        fields: Dict[str, BaseField] = {}
        model_references: Dict[str, "ModelRef"] = {}
        managers: Dict[str, BaseManager] = {}
        meta_class: "object" = attrs.get("Meta", type("Meta", (), {}))
        base_annotations: Dict[str, Any] = {}
        has_explicit_primary_key = False
        registry: Optional[Registry] = get_model_registry(bases, meta_class)
        is_abstract: bool = getattr(meta_class, "abstract", False)
        parents = [parent for parent in bases if isinstance(parent, BaseModelMeta)]

        # Extract the custom Edgy Fields in a pydantic format.
        attrs, model_fields = extract_field_annotations_and_defaults(attrs)

        # Extract fields and managers and include them in attrs
        attrs = extract_fields_and_managers(bases, attrs)

        for key, value in attrs.items():
            if isinstance(value, BaseField):
                if key == "pk" and not isinstance(value, PKField):
                    raise ImproperlyConfigured(
                        f"Cannot add a field named pk to model {name}. Protected name."
                    )
                # make sure we have a fresh copy where we can set the owner
                value = copy.copy(value)
                if value.primary_key:
                    has_explicit_primary_key = True
                # set as soon as possible the field_name
                value.name = key
                if registry:
                    value.registry = registry

                # add fields and non BaseRefForeignKeyField to fields
                # The BaseRefForeignKeyField is actually not a normal SQL Foreignkey
                # It is an Edgy specific operation that creates a reference to a ForeignKey
                # That is why is not stored as a normal FK but as a reference but
                # stored also as a field to be able later or to access anywhere in the model
                # and use the value for the creation of the records via RefForeignKey.
                # This is then used in `save_model_references()` and `update_model_references
                # saving a reference foreign key.
                # We split the keys (store them) in different places to be able to easily maintain and
                #  what is what.
                if isinstance(value, BaseRefForeignKeyField):
                    model_references[key] = value.to
                else:
                    fields[key] = value
            elif isinstance(value, BaseManager):
                value = copy.copy(value)
                value.name = key
                managers[key] = value

        if not is_abstract:
            # the order is important because it reflects the inheritance order
            # from the strongest to the weakest
            fieldnames_to_check = deque(fields.keys())
            while fieldnames_to_check:
                # pop the weakest next field
                field_name = fieldnames_to_check.pop()
                field = fields[field_name]
                embedded_fields = field.get_embedded_fields(field_name, fields)
                if embedded_fields:
                    for sub_field_name, sub_field in embedded_fields.items():
                        if sub_field_name == "pk":
                            raise ValueError("sub field uses reserved name pk")

                        if sub_field_name in fields and fields[sub_field_name].owner is None:
                            raise ValueError(f"sub field name collision: {sub_field_name}")
                        # set as soon as possible the field_name
                        sub_field.name = sub_field_name
                        if registry:
                            sub_field.registry = registry
                        if sub_field.primary_key:
                            has_explicit_primary_key = True
                        # insert as stronger element
                        fieldnames_to_check.appendleft(sub_field_name)
                        fields[sub_field_name] = sub_field
                        model_fields[sub_field_name] = sub_field
            # Handle with multiple primary keys and auto generated field if no primary key is provided
            if not is_abstract and parents and registry and not has_explicit_primary_key:
                if "id" not in fields:
                    if attrs.get("__reflected__", False):
                        raise ImproperlyConfigured(
                            f"Cannot create model {name}. No primary key found and reflected."
                        )
                    else:
                        fields["id"] = edgy_fields.BigIntegerField(
                            primary_key=True, autoincrement=True, inherit=False, name="id"
                        )  # type: ignore
                if not isinstance(fields["id"], BaseField) or not fields["id"].primary_key:
                    raise ImproperlyConfigured(
                        f"Cannot create model {name} without explicit primary key if field 'id' is already present."
                    )

        for field_name in fields:
            attrs.pop(field_name, None)

        for manager_name in managers:
            attrs.pop(manager_name, None)

        attrs["meta"] = meta = MetaInfo(meta_class, fields_mapping=fields, model_references=model_references, parents=parents, managers=managers)
        if is_abstract:
            meta.abstract = True

        del is_abstract
        if not fields:
            meta.abstract = True

        if not meta.abstract:
            fields["pk"] = PKField(
                exclude=True,
                name="pk",
                inherit=False
            )

        model_class = super().__new__

        # Handle annotations
        annotations: Dict[str, Any] = handle_annotations(bases, base_annotations, attrs)

        for k, _ in meta.managers.items():
            if annotations and k not in annotations:
                raise ImproperlyConfigured(
                    f"Managers must be type annotated and '{k}' is not annotated. Managers must be annotated with ClassVar."
                )
            if annotations and get_origin(annotations[k]) is not ClassVar:
                raise ImproperlyConfigured("Managers must be ClassVar type annotated.")

        # Ensure the initialization is only performed for subclasses of EdgyBaseModel
        attrs["__init_annotations__"] = annotations

        # Ensure initialization is only performed for subclasses of EdgyBaseModel
        # (excluding the EdgyBaseModel class itself).
        if not parents:
            return model_class(cls, name, bases, attrs)

        new_class = cast("Type[Model]", model_class(cls, name, bases, attrs))
        new_class.fields = fields

        # Update the model_fields are updated to the latest
        new_class.model_fields = {**new_class.model_fields, **model_fields}

        # Set the owner of the field, must be done as early as possible
        # don't use meta.fields_mapping to not trigger the lazy evaluation
        for value in fields.values():
            value.owner = new_class
        # set the model_class of managers
        for value in meta.managers.values():
            value.owner = new_class


        # Validate meta for uniques and indexes
        if meta.abstract:
            if getattr(meta, "unique_together", None) is not None:
                raise ImproperlyConfigured("unique_together cannot be in abstract classes.")

            if getattr(meta, "indexes", None) is not None:
                raise ImproperlyConfigured("indexes cannot be in abstract classes.")

        # Now set the registry of models
        if meta.registry is None:
            if getattr(new_class, "__db_model__", False):
                meta.registry = registry
            else:
                new_class.model_rebuild(force=True)
                return new_class
        if registry is None:
            raise ImproperlyConfigured(
                "Registry for the table not found in the Meta class or any of the superclasses. You must set the registry in the Meta."
            )


        new_class.database = registry.database

        # Making sure the tablename is always set if the value is not provided
        if getattr(meta, "tablename", None) is None:
            tablename = f"{name.lower()}s"
            meta.tablename = tablename

        if getattr(meta, "unique_together", None) is not None:
            unique_together = meta.unique_together
            if not isinstance(unique_together, (list, tuple)):
                value_type = type(unique_together).__name__
                raise ImproperlyConfigured(
                    f"unique_together must be a tuple or list. Got {value_type} instead."
                )
            else:
                for value in unique_together:
                    if not isinstance(value, (str, tuple, UniqueConstraint)):
                        raise ValueError(
                            "The values inside the unique_together must be a string, a tuple of strings or an instance of UniqueConstraint."
                        )

        # Handle indexes
        if getattr(meta, "indexes", None) is not None:
            indexes = meta.indexes
            if not isinstance(indexes, (list, tuple)):
                value_type = type(indexes).__name__
                raise ImproperlyConfigured(
                    f"indexes must be a tuple or list. Got {value_type} instead."
                )
            else:
                for value in indexes:
                    if not isinstance(value, Index):
                        raise ValueError("Meta.indexes must be a list of Index types.")

        for value in fields.values():
            if isinstance(value, BaseManyToManyForeignKeyField):
                value.create_through_model()

        # Making sure it does not generate tables if abstract it set
        if not meta.abstract:
            if getattr(cls, "__reflected__", False):
                registry.reflected[name] = new_class
            else:
                registry.models[name] = new_class

        new_class.__db_model__ = True
        meta.model = new_class

        # Sets the foreign key fields
        if not new_class.is_proxy_model and meta.foreign_key_fields:
            _set_related_name_for_foreign_keys(meta.foreign_key_fields, new_class)

        # Update the model references with the validations of the model
        # Being done by the Edgy fields instead.
        # Generates a proxy model for each model created
        # Making sure the core model where the fields are inherited
        # And mapped contains the main proxy_model
        if not new_class.is_proxy_model and not meta.abstract:
            proxy_model = new_class.generate_proxy_model()
            new_class.__proxy_model__ = proxy_model
            new_class.__proxy_model__.parent = new_class
            new_class.__proxy_model__.model_rebuild(force=True)
            meta.registry.models[new_class.__name__] = new_class  # type: ignore

        # finalize
        new_class.model_rebuild(force=True)

        return new_class

    def get_db_shema(cls) -> Union[str, None]:
        """
        Returns a db_schema from registry if any is passed.
        """
        if hasattr(cls, "meta") and hasattr(cls.meta, "registry"):
            return cast(str, cls.meta.registry.db_schema)
        elif hasattr(cls, "__using_schema__") and cls.__using_schema__ is not None:
            return cast(str, cls.__using_schema__)
        return None

    @property
    def table(cls) -> Any:
        """
        Making sure the tables on inheritance state, creates the new
        one properly.

        Making sure the following scenarios are met:

        1. If there is a context_db_schema, it will return for those, which means, the `using`
        if being utilised.
        2. If a db_schema in the `registry` is passed, then it will use that as a default.
        3. If none is passed, defaults to the shared schema of the database connected.
        """
        if not hasattr(cls, "_table"):
            cls._table = cls.build(cls.get_db_shema())
        elif hasattr(cls, "_table"):
            table = cls._table
            if table.name.lower() != cls.meta.tablename:
                cls._table = cls.build(cls.get_db_shema())
        return cls._table

    @table.setter
    def table(cls, value: sqlalchemy.Table) -> None:
        try:
            del cls._pknames
        except AttributeError:
            pass
        try:
            del cls._pkcolumns
        except AttributeError:
            pass
        cls._table = value

    def __getattr__(cls, name: str) -> Any:
        """
        For controlling inheritance we cannot have manager descriptors, so fake them
        """
        manager = cls.meta.managers.get(name)
        if manager is not None:
            return manager
        return super().__getattr__(name)


    @property
    def pkcolumns(cls) -> Sequence[str]:
        if cls.__dict__.get("_pkcolumns", None) is None:
            build_pkcolumns(cls)
        return cast(Sequence[str], cls._pkcolumns)

    @property
    def pknames(cls) -> Sequence[str]:
        if cls.__dict__.get("_pknames", None) is None:
            build_pknames(cls)
        return cast(Sequence[str], cls._pknames)

    @property
    def signals(cls) -> signals_module.Broadcaster:
        """
        Returns the signals of a class
        """
        meta: MetaInfo = cls.meta
        return meta.signals

    def table_schema(cls, schema: str) -> Any:
        """
        Making sure the tables on inheritance state, creates the new
        one properly.

        The use of context vars instead of using the lru_cache comes from
        a warning from `ruff` where lru can lead to memory leaks.
        """
        return cls.build(schema=schema)

    @property
    def proxy_model(cls) -> Any:
        """
        Returns the proxy_model from the Model when called using the cache.
        """
        return cls.__proxy_model__

    @property
    def columns(cls) -> sqlalchemy.sql.ColumnCollection:
        return cast("sqlalchemy.sql.ColumnCollection", cls.table.columns)



from typing import ClassVar

from edgy.contrib.multi_tenancy.metaclasses import BaseTenantMeta, TenantMeta
from edgy.core.db.models.model import Model


class TenantModel(Model, metaclass=BaseTenantMeta):
    """
    Base for a multi tenant model from the Edgy contrib.
    This is **not mandatory** and can be used as a possible
    out of the box solution for multi tenancy.

    This design is not meant to be "the one" but instead an
    example of how to achieve the multi-tenancy in a simple fashion
    using Edgy and Edgy models.
    """

    meta: ClassVar[TenantMeta] = TenantMeta(None, abstract=True)



from typing import TYPE_CHECKING, Any, Optional, Type, Union, cast

from edgy.core.db.context_vars import get_tenant, set_tenant
from edgy.core.db.querysets.base import QuerySet

if TYPE_CHECKING:
    from edgy.core.db.models.base import EdgyBaseModel

class BaseManager:
    def __init__(self, *, owner: Optional[Union[Type["EdgyBaseModel"]]] = None, inherit: bool=True, name: str = "", instance: Optional[Union["EdgyBaseModel"]]=None):
        self.owner = owner
        self.inherit = inherit
        self.name = name
        self.instance = instance

    @property
    def model_class(self) -> Any:
        # legacy name
        return self.owner

    def get_queryset(self) -> QuerySet:
        """
        Returns the queryset object.
        """
        raise NotImplementedError(f"The {self!r} manager doesn't implement the get_queryset method.")


class Manager(BaseManager):
    """
    Base Manager for the Edgy Models.
    To create a custom manager, the best approach is to inherit from the ModelManager.

    **Example**

    ```python
    from saffier.managers import ModelManager
    from saffier.models import Model


    class MyCustomManager(ModelManager): ...


    class MyOtherManager(ModelManager): ...


    class MyModel(saffier.Model):
        query = MyCustomManager()
        active = MyOtherManager()

        ...
    ```
    """

    def get_queryset(self) -> "QuerySet":
        """
        Returns the queryset object.

        Checks for a global possible tenant and returns the corresponding queryset.
        """
        tenant = get_tenant()
        if tenant:
            set_tenant(None)
            return QuerySet(
                self.owner,
                table=self.owner.table_schema(tenant),  # type: ignore
            )
        return QuerySet(self.owner)

    def __getattr__(self, name: str) -> Any:
        """
        Gets the attribute from the queryset and if it does not
        exist, then lookup in the model.
        """
        if name.startswith("_") or name == self.name:
            return super().__getattr__(name)
        try:
            # we need to check tenant every request
            return getattr(self.get_queryset(), name)
        except AttributeError:
            return getattr(self.owner, name)


class RedirectManager(BaseManager):
    def __init__(self, *, redirect_name: str, **kwargs: Any):
        self.redirect_name = redirect_name
        super().__init__(**kwargs)

    def __getattr__(self, name: str) -> Any:
        if name.startswith("_") or name == self.name:
            return super().__getattr__(name)
        return getattr(self.owner.meta.managers[self.redirect_name], name)  # type: ignore

    def get_queryset(self) -> "QuerySet":
        return cast("QuerySet", self.__getattr__("get_queryset")())



import click

from edgy.cli.base import heads as _heads
from edgy.cli.env import MigrationEnv


@click.option(
    "-d",
    "--directory",
    default=None,
    help=('Migration script directory (default is "migrations")'),
)
@click.option("-v", "--verbose", is_flag=True, help="Use more verbose output")
@click.option("--resolve-dependencies", is_flag=True, help="Treat dependency versions as down revisions")
@click.command()
def heads(env: MigrationEnv, directory: str, verbose: bool, resolve_dependencies: bool) -> None:
    """Show current available heads in the script directory"""
    _heads(env.app, directory, verbose, resolve_dependencies)



from typing import TYPE_CHECKING, Dict, Type, Union

import sqlalchemy
from loguru import logger

from edgy.core.terminal import Terminal
from edgy.exceptions import ModelSchemaError

terminal = Terminal()

if TYPE_CHECKING:
    from edgy import Model, Registry


def table_schema(model_class: Type["Model"], schema: str) -> sqlalchemy.Table:
    """
    Making sure the tables on inheritance state, creates the new
    one properly.

    The use of context vars instead of using the lru_cache comes from
    a warning from `ruff` where lru can lead to memory leaks.
    """
    return model_class.build(schema)


async def create_tables(registry: "Registry", models: Dict[str, Type["Model"]], schema: str) -> None:
    """
    Creates the table models for a specific schema just generated.

    Iterates through the tenant models and creates them in the schema.
    """

    for name, model in models.items():
        table = table_schema(model, schema)

        logger.info(f"Creating table '{name}' for schema: '{schema}'")
        try:
            async with registry.engine.begin() as connection:
                await connection.run_sync(table.create)
            await registry.engine.dispose()
        except Exception as e:
            logger.error(str(e))
            ...


async def create_schema(
    registry: "Registry",
    schema_name: str,
    models: Union[Dict[str, "Model"], None] = None,
    if_not_exists: bool = False,
    should_create_tables: bool = False,
) -> None:
    """
    Creates a schema in a given registry.

    This function creates a new schema for a tenant in the provided registry.
    It optionally checks if the schema already exists and creates tables within
    the schema if specified.

    Parameters:
        registry (Registry): The registry object where the schema will be created.
        schema_name (str): The name of the schema to be created.
        models (dict[str, Model], optional): When models is provided, it will use them to generate
        those same models in a given schema or else it will default to the `registry.models`.
        if_not_exists (bool, optional): If True, the schema will be created only if it does not already exist.
        Defaults to False.
        should_create_tables (bool, optional): If True, tables will be created within the new schema. Defaults to False.

    Raises:
        ModelSchemaError: If the schema name is the same as the default schema name.
    """
    default_schema_name = registry.schema.get_default_schema() or "public"
    if schema_name.lower() == default_schema_name.lower():
        raise ModelSchemaError(f"Cannot create a schema with the same name as the default: '{schema_name}'.")

    # Create the new schema, optionally checking if it already exists
    await registry.schema.create_schema(schema_name, if_not_exists=if_not_exists)

    # Optionally create tables within the new schema
    if should_create_tables:
        terminal.write_info(f"Creating the tables in schema: {schema_name}")
        schema_models = models or registry.models
        await create_tables(registry, schema_models, schema_name)



import typing
from inspect import isclass
from typing import TYPE_CHECKING, TypeVar

from typing_extensions import get_origin

import edgy
from edgy.core.db.fields.base import BaseField, BaseForeignKey
from edgy.core.db.fields.factories import ForeignKeyFieldFactory
from edgy.core.terminal import Print
from edgy.exceptions import ModelReferenceError

if TYPE_CHECKING:

    from edgy import Model
    from edgy.core.db.models.model_reference import ModelRef

T = TypeVar("T", bound="Model")


terminal = Print()


class BaseRefForeignKeyField(BaseForeignKey):
    pass


class RefForeignKey(ForeignKeyFieldFactory, list):
    _bases = (BaseRefForeignKeyField,)
    _type = list

    @classmethod
    def is_class_and_subclass(cls, value: typing.Any, _type: typing.Any) -> bool:
        original = get_origin(value)
        if not original and not isclass(value):
            return False

        try:
            if original:
                return original and issubclass(original, _type)
            return issubclass(value, _type)
        except TypeError:
            return False

    def __new__(cls, to: "ModelRef", null: bool = False) -> BaseField:  # type: ignore
        if not cls.is_class_and_subclass(to, edgy.ModelRef):
            raise ModelReferenceError(detail="A model reference must be an object of type ModelRef")
        if not hasattr(to, "__model__") or getattr(to, "__model__", None) is None:
            raise ModelReferenceError("'__model__' must bre declared when subclassing ModelRef.")

        return super().__new__(cls, to=to, null=null)
